from pyrateshield import labels
from pyrateshield.modelitem import ModelItem
from pyrateshield.dosemapper import Dosemapper, get_critical_point

from pyrateshield.constants import CONSTANTS
from PyQt5.QtCore import QPointF
from PyQt5.QtGui import QPolygonF

import math
def ZERO_VERTICES():
    return [[0, 0], [0, 0]]

def ZERO_POSITION():
   return [0, 0]      

def EMPTY_MATERIALS():
    # default for shielding
    return [[labels.EMPTY_MATERIAL, 0], 
            [labels.EMPTY_MATERIAL, 0]]

class PositionModelItem(ModelItem):
    def setPositionPixels(self, pos):
        pos *= self.get_pixel_size_cm()
        self.position = [pos.x(), pos.y()]
        
    def positionPixels(self):
        return QPointF(*self.position) / self.get_pixel_size_cm()
    
    def move_pixels(self, dx, dy):
        pos_pixels = self.positionPixels()
        new_pos = QPointF(pos_pixels.x() + dx, pos_pixels.y() + dy)
        self.setPositionPixels(new_pos)
    

class Ruler(ModelItem):
    default_name = 'ruler'
    label = labels.RULERS
    
    _attr_dct = {"vertices_pixels": labels.VERTICES_PIXELS}
    _attr_defaults = {labels.VERTICES_PIXELS: ZERO_VERTICES}
    
        
    def tooltip(self):
        return f'Distance: {self.length_cm} cm'
        
    @property
    def vertices(self):
        psize = self.get_pixel_size_cm()
        return [[vi*psize for vi in vii] for vii in self.vertices_pixels]
    
    @vertices.setter
    def vertices(self, vertices):
        psize = self.get_pixel_size_cm()
        vp =  [[vi/psize for vi in vii] for vii in vertices]
        self.vertices_pixels = vp
    
    @property
    def length_pixels(self):
        p1, p2 = self.vertices_pixels
        x1, y1 = p1
        x2, y2 = p2
        return ((x2-x1)**2 + (y2-y1)**2)**0.5
        
    @property
    def length_cm(self):
        return self.length_pixels * self.get_pixel_size_cm()
    
    def move_pixels(self, dx, dy):
        new_vp = []
        for vp in self.vertices_pixels:
            new_vp.append([vp[0]+dx, vp[1]+dy])
        self.vertices_pixels=new_vp
        
    @property
    def angle(self):
         vvp = self.vertices_pixels
         dx = vvp[1][0] - vvp[0][0]
         dy = vvp[1][1] - vvp[0][1]
         return math.degrees(math.atan2(dy, dx))
     
    def isClosed(self):
         return False
        

class Material(ModelItem):
    # buildup table and attenuation table are used by pyshield only
    default_name = 'material'
    label = labels.MATERIALS
    _attr_dct = {
        "name": labels.NAME,
        "density": labels.DENSITY,
        "attenuation_table": labels.ATTENUATION_TABLE,
        "buildup_table": labels.BUILDUP_TABLE,
        "radtracer_material": labels.RADTRACER_MATERIAL
    }
    
    _attr_defaults = {labels.NAME: default_name,
                      labels.DENSITY: 1,
                      labels.ATTENUATION_TABLE: labels.EMPTY_TABLE,
                      labels.BUILDUP_TABLE: labels.EMPTY_TABLE,
                      labels.RADTRACER_MATERIAL: labels.EMPTY_MATERIAL}
                      
    
class Shielding(ModelItem):
    default_name = 'shielding'
    label = labels.SHIELDINGS

    _attr_dct = {
        "name": labels.NAME,
        "color": labels.COLOR,
        "linewidth": labels.LINEWIDTH,
        "materials": labels.MATERIALS,
        
    }
    
    _attr_defaults = {labels.NAME: default_name,
                      labels.COLOR: 'black',
                      labels.LINEWIDTH: 2,
                      labels.MATERIALS: EMPTY_MATERIALS}
    
                 
   
    
    
class Clearance(ModelItem):
    default_name = 'clearance'
    label = labels.CLEARANCE
    
    _attr_dct = {
        "name": labels.NAME,
        'apply_fraction1': labels.APPLY_FRACTION1,
        'apply_fraction2': labels.APPLY_FRACTION2,
        "fraction1": labels.DECAY_FRACTION1,
        "fraction2": labels.DECAY_FRACTION2,
        "half_life1": labels.HALFLIFE1,
        'half_life2': labels.HALFLIFE2,    
        'apply_split_fractions': labels.ENABLE_SPLIT_FRACTIONS,
        'split_time': labels.SPLIT_FRACTION_TIME}
       
    
    
    _attr_defaults = {labels.NAME: default_name,
                      labels.APPLY_FRACTION1: False,
                      labels.DECAY_FRACTION1: 0,
                      labels.HALFLIFE2: float('Inf'),
                      labels.APPLY_FRACTION2: False,
                      labels.DECAY_FRACTION2: 0,
                      labels.HALFLIFE1: float('Inf'),          
                      labels.ENABLE_SPLIT_FRACTIONS: False,
                      labels.SPLIT_FRACTION_TIME: 24}
    
            
    
    def is_equal(self, other):
        if isinstance(other, Clearance):
            eq = True
            for attr in self._attr_dct.keys():
                if attr == 'name': continue
                eq = eq and (getattr(self, attr) == getattr(other, attr))
            return eq
        else:
            return False
        
    
    # LEGACY, Old psp files use single half_life
    @classmethod
    def from_half_life(cls, half_life):
        return cls(fraction1=1,
                   half_life1=half_life)
        


class Wall(PositionModelItem):
    label = labels.WALLS

    _attr_dct = {
        "vertices": labels.VERTICES, 
        "shielding": labels.SHIELDING,
        "closed": labels.CLOSED
    }
    
    _attr_defaults = {
        labels.VERTICES: ZERO_VERTICES,
        labels.SHIELDING: labels.EMPTY_SHIELDING,
        labels.CLOSED: False
    }
    
    def __len__(self):
        return len(self.vertices)
    
    def isClosed(self):
        return self.closed and len(self) > 2
    
    def isClosable(self):
        return not self.closed and len(self) > 2
    
    def close(self):
        self.closed=True
        
        
    def open(self, index):
       
        if index < 0 or index < len(self.vertices) - 1:
            self.vertices = self.vertices[index+1:] + self.vertices[0:index+1]
            
        self.closed = False
        
    def positionPixels(self):                
        return self.vertices2poly(self.vertices, self.get_pixel_size_cm())
    def setPositionPixels(self, poly):
        self.vertices = self.poly2vertices(poly, self.get_pixel_size_cm())
        
    def insertVertexPixelPoint(self, index, vertex):
        self.insertVertex(index, [vertex.x()*self.get_pixel_size_cm(),
                                  vertex.y()*self.get_pixel_size_cm()])
        
    def move_pixels(self, dx, dy):
        poly = self.positionPixels()
        moved_poly = QPolygonF()
        for i in range(poly.size()):
            moved_poly.append(poly[i]+QPointF(dx, dy))
            
        self.setPositionPixels(moved_poly)
            
        
    def insertVertex(self, index, vertex):
        self.valid_vertex(vertex)
        old_value = [[vi for vi in vii] for vii in self.vertices]
        self.vertices.insert(index, vertex)
        
        self.update_event.emit([self, labels.VERTICES, old_value, self.vertices])
        
    def deleteVertex(self, index):  
        if len(self) == 2:
            raise IndexError()
        old_value = [[vi for vi in vii] for vii in self.vertices]
        self.vertices.pop(index)
        if len(self) == 2:
            self.closed = False            
        self.update_event.emit([self, labels.VERTICES, old_value, self.vertices])
        
    def setVertexPixelPoint(self, index, vertex):
        self.setVertex(index, [vertex.x()*self.get_pixel_size_cm(),
                               vertex.y()*self.get_pixel_size_cm()])
        
    def closestVertexToPixelPoint(self, point):
        distance = [((pi.x() - point.x())**2 + (pi.y() - point.y())**2)** 0.5\
                    for pi in self.positionPixels()]
        index = distance.index(min(distance))
        return index, self.positionPixels()[index], min(distance)
            
        
    def setVertex(self, index, vertex):
        old_value = [[vi for vi in vii] for vii in self.vertices]
        self.valid_vertex(vertex)            
        self.vertices[index] = vertex 
        self.update_event.emit([self, labels.VERTICES, old_value, self.vertices])
        
    def equal_vertices(vertices1, vertices2):
        if vertices1 is None and vertices2 is None:
            return True
        elif vertices1 is None or vertices2 is None:
            return False
        
        
        Wall.valid_vertices(vertices1)
        Wall.valid_vertices(vertices2)
        
        equal = True
        
        if len(vertices1) != len(vertices2):
            return False
        
        for vi, vii in zip(vertices1, vertices2):
            if vi[0] != vii[0] or vi[1] != vii[1]:
                equal = False
        return equal
    
    def get_shielding(self):
        if self.parent() is None:
            return None
        else:
            return self.parent().parent().shieldings.itemByName(self.shielding)
    
    @staticmethod
    def poly2vertices(poly, pixel_size_cm=1):
        return [[point.x()*pixel_size_cm, point.y()*pixel_size_cm] for point in poly]
    
    @staticmethod
    def vertices2poly(vertices, pixel_size_cm=1):
        return QPolygonF([QPointF(vi[0], vi[1])/pixel_size_cm for vi in vertices])
    
    @staticmethod
    def valid_vertex(vertex, raise_error=True):
        if isinstance(vertex, list) and len(vertex) == 2\
            and all([isinstance(value, (int, float)) for value in vertex]):
                return True   

        
        return False
    
    @staticmethod
    def valid_vertices(vertices, raise_error=True):
        if isinstance(vertices, list) and\
            all([Wall.valid_vertex(vi, raise_error=False) for vi in vertices]):
                return True
            
        if raise_error:
            msg = f'Invalid vertices of type {type(vertices)} and value {vertices}'
            raise ValueError(msg)
            
        return False
                

        
class CriticalPoint(PositionModelItem):
    default_name = 'critical point'
    label = labels.CRITICAL_POINTS

    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION, 
        "occupancy_factor": labels.OCCUPANCY_FACTOR,
        'enabled': labels.ENABLED
    }
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.OCCUPANCY_FACTOR: 1,
        labels.ENABLED: True
    }
    
    def tooltip(self):
        dose_pyshield = round(self.dose_pyshield(), 3)
        dose_radtracer = round(self.dose_radtracer(), 3)
        of = self.occupancy_factor
        dose_pyshield_of = round(dose_pyshield*of, 3)
        dose_radtracer_of = round(dose_radtracer*of, 3)
        
        if dose_pyshield == 0:
            dose_pyshield = "<1E-3"
            
        if dose_radtracer == 0:
            dose_radtracer = "<1E-3"
            
        if dose_radtracer_of == 0:
            dose_radtracer_of = "<1E-3"
            
        if dose_pyshield_of == 0:
            dose_pyshield_of = "<1E-3"

        text = super().tooltip()
        text += "Dose [mSv]:\n"
        text += f"\tpyshield: {dose_pyshield}\n"
        text += f"\tradtracer: {dose_radtracer}\n"
        text += "Dose corrected for occupancy [mSv]:\n"
        text += f"\tpyshield: {dose_pyshield_of}\n"
        text += f"\tradtracer: {dose_radtracer_of}"
        return text
    
    def dose_pyshield(self, occupancy_correction=False):
        dose = get_critical_point(self.parent().parent(), self,
                                  engine=labels.PYSHIELD, 
                                  sum_dose=True)
        if occupancy_correction:
            dose *= self.occupancy_factor
        return dose
    
    def dose_radtracer(self, occupancy_correction=False):
        dose = get_critical_point(self.parent().parent(), self,
                                  engine=labels.RADTRACER, 
                                  sum_dose=True)
        if occupancy_correction:
            dose *= self.occupancy_factor
        return dose
    

class SourceNM(PositionModelItem):
    default_name = 'source NM'
    label = labels.SOURCES_NM
    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION,  
        "number_of_exams": labels.NUMBER_OF_EXAMS,
        "isotope": labels.ISOTOPE,
        "self_shielding": labels.SELF_SHIELDING,
        "activity": labels.ACTIVITY,
        "duration": labels.DURATION,
        "apply_decay_correction": labels.APPLY_DECAY_CORRECTION,
        'clearance': labels.CLEARANCE,
        'enabled': labels.ENABLED
    }
    
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.NUMBER_OF_EXAMS: 1,
        labels.ISOTOPE: 'F-18',
        labels.SELF_SHIELDING: 'None', # must be str 'None' or 'Body' 
        labels.ACTIVITY: 1,
        labels.DURATION: 1,
        labels.APPLY_DECAY_CORRECTION: True,
        labels.CLEARANCE: labels.EMPTY_CLEARANCE,
        labels.ENABLED: True}
                

    
    @property
    def tiac(self):
        return self.time_integrated_activity_coefficient_mbqh

class SourceXray(PositionModelItem):
    default_name = 'source Xray'
    label = labels.SOURCES_XRAY
    _kvp = None
    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION, 
        "number_of_exams": labels.NUMBER_OF_EXAMS,
        "kvp": labels.KVP,
        "dap": labels.DAP,
        "enabled": labels.ENABLED,
    }
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.DAP: 1,
        labels.NUMBER_OF_EXAMS: 1,
        labels.ENABLED: True,
        labels.KVP: 100 #CONSTANTS.xray[0].kvp
        }

        

class SourceCT(PositionModelItem):
    default_name = 'source CT'
    label = labels.SOURCES_CT
    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION, 
        "number_of_exams": labels.NUMBER_OF_EXAMS,
        "body_part": labels.CT_BODY_PART,
        "kvp": labels.KVP,
        "dlp": labels.DLP,
        'enabled': labels.ENABLED
    }
    
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.DLP: 1,
        labels.NUMBER_OF_EXAMS: 1,
        labels.CT_BODY_PART: 'Body',
        labels.KVP: 120, #CONSTANTS.ct[0].kvp,
        labels.ENABLED: True}
        
    
    @property
    def available_kvp(self):
        return list(set([item.kvp for item in CONSTANTS.ct]))
    
   




import time
from pyrateshield import shared, labels
from pyrateshield.constants import CONSTANTS
from pyrateshield.pyshield import isotope, doserates

import math
import numpy as np
import timeit


CUT_OFF_KEV = 30

def add_dict(dict1, dict2):
    sum_dict = dict1.copy()
    for key, value in dict2.items():
        if key in sum_dict.keys():
            sum_dict[key] += value
        else:
            sum_dict[key] = value
    return sum_dict

class H10:
    _values = {}
    @classmethod
    def h10(cls, keV, abundance):
        if (keV, abundance) not in cls._values.keys():
            value = doserates.H10(keV, abundance=abundance)
            cls._values[(keV, abundance)] = value
        return cls._values[(keV, abundance)] 
        

class GammaRayOnGrid:
    _h10 = None
    _distance_map = None
    _dose_map = None
    
    def __init__(self, position, keV=0, abundance=0, grid=None):
        self.x, self.y = float(position[0]), float(position[1])
        self.keV = keV
        self.abundance = abundance
        self.grid = grid
        self._cache = {}
            
    @property
    def h10(self):
        # uSv/ h per MBq / m2
        return H10.h10(self.keV, abundance=self.abundance)
        

    def dose_at_point(self, point):
        # mSv
        x, y = point
        distance = np.sqrt((x-self.x)**2 + (y-self.y)**2) / 100 # cm --> m
        if distance == 0:
            return float('Inf')
        else:
            dose = self.h10 / (distance**2) / 1000 #uSv --> mSv
        return dose

    @property
    def dosemap(self):
        return self.get_dosemap()

    def get_dosemap(self):
        # mSv / h per MBq / m2
        distance_map = self.grid.distance_map_meters(self.x, self.y)
        return self.h10 /  distance_map**2 / 1000 # uSv --> mSv
    
    

class IsotopeOnGrid:
    _gammarays = None
    _cache = None
    def __init__(self, grid=None, position=None, isotope_name=None):
        self.isotope_name = isotope_name
        self.position = position
        self.grid = grid
       

    def dose_at_point(self, point):
        
        dose = 0
        for gammaray in self.gammarays():
            dose += gammaray.dose_at_point(point)
           
        return dose

    @property
    def isotope(self):
        return CONSTANTS.get_isotope_by_name(self.isotope_name)
    
    @property
    def decay_constant(self):
        return np.log(2) / self.isotope.half_life
    
    @property
    def keVs(self):
        return [item[0] for item in self.isotope.spectrum if item[0] > CUT_OFF_KEV]
    
    @property
    def abundance(self):
        return [item[1]/100 for item in self.isotope.spectrum if item[0] > CUT_OFF_KEV]
    
        
    @property
    def gammarays(self):
        if self._gammarays is None:
            self._gammarays = {}
            for ei, ai in zip(self.keVs, self.abundance):
                ray = GammaRayOnGrid(grid=self.grid, keV=ei, abundance=ai, 
                                     position=self.position)
                self._gammarays[ei] = ray
                                                   
        return self._gammarays

    
class SourceWallMap():
    _intersection_map = None
    _material_map = None
    _intersection_time = 0
    _counter = 0
    def __init__(self, source=None, wall=None, dosemap=None, shielding=None,
                 materials=None):
        self.materials = materials
        self.source = source
        self.wall = wall
        self.shielding = shielding
        self.dosemap = dosemap

    @property
    def intersection_map(self):
       
        start = time.time()
        if self._intersection_map is None:
            point = self.source.position
            vertices = self.wall.vertices
            poly = shared.get_polygon(point, vertices, self.dosemap.extent)
            
            points, mask = shared.grid_points_in_polygon(self.dosemap.grid,
                                                         poly)
            
            intersection_map = shared.get_intersection_map(points, point, 
                                              vertices, mask)
            
            self._intersection_map = intersection_map
        stop = time.time()
        self._intersection_time += (stop-start)
        return self._intersection_map
    
    @property
    def material_map(self):
       
        if self._material_map is None:
            material_map = {}
            for material_name, thickness in self.shielding.materials:
                material_map[material_name] = (thickness * self.intersection_map)             
            self._material_map = material_map
            
        return self._material_map
    
    def line_intersect(self, point):
       
        if shared.intersect(*self.wall.vertices, self.source.position, point):
            return shared.get_diagonal_intersection_factor(point, self.source.position, self.wall.vertices)
        else:
             return None
            
    def line_intersect_materials(self, point):
        diag_factor = self.line_intersect(point)
        intersection_materials = {}
        
        if diag_factor:
            for material_name, thickness in self.shielding.materials:                
                intersection_materials[material_name] = thickness * diag_factor
                
            intersection_materials.pop(labels.EMPTY_MATERIAL, None)        
            
        return intersection_materials

            
class SourceWallsMap():
    _gammarays = None
    _material_map = None
    _attenuation_map = None
    _buildup_map = None
    _isotope = None
    
    
    
    def __init__(self, dosemap=None, source=None, walls=None, shieldings=None, 
                 clearance=None, materials=None, parent=None):
        
        self.source = source
        self.clearance = clearance
        self.walls = walls
        self.shieldings = shieldings 
        self._items = []
        self.dosemap=dosemap
        self.materials = materials
        self.parent = parent
        self.intersections = {}
        
        

        for wall in walls:
            shielding = shieldings.itemByName(wall.shielding)
            sw_map = SourceWallMap(source=source, wall=wall, dosemap=dosemap,
                                   shielding=shielding)
            self._items += [sw_map]
    
    @property
    def self_shielding_materials(self):
        if self.source.self_shielding == 'Body':            
            material_name = CONSTANTS.self_shielding_pyshield[labels.MATERIAL]
            thickness = CONSTANTS.self_shielding_pyshield[labels.THICKNESS]            
            return {material_name: thickness}
                    
        else:
            return {}
        
    @property
    def self_shielding_map(self):
        material_map = {}
        shielding_map = np.ones(self.dosemap.grid.X.shape)
        for material, thickness in self.self_shielding_materials.items():
            
            material_map = add_dict(material_map, {material: thickness * shielding_map})
        return material_map
            
        
    def line_intersect_materials(self, point):
        if point not in self.intersections.keys():
            materials = {}
            for item in self._items:
                materials = add_dict(materials, item.line_intersect_materials(point))
                
            materials = add_dict(materials, self.self_shielding_materials) 
            
            materials = {k: v for k, v in materials.items()\
                         if k != labels.EMPTY_MATERIAL}
            self.intersections[point] = materials
        return self.intersections[point]
    
 
    def dose_at_point(self, point):
        point = tuple(point)
        # SourceWallsMap._time
        # start = time.time()
        materials = self.line_intersect_materials(point)
        # stop = time.time()
        # SourceWallsMap._time += (stop - start)
       
        
        sumdose = 0
        for gammaray in self.isotope.gammarays.values():
            
            dose = gammaray.dose_at_point(point)
           

            for material_name, thickness in materials.items():
                material = self.materials.itemByName(material_name)
                
                
                
                attenuation = isotope.AttenuationHelper.calculate(material,
                                                                  gammaray.keV,
                                                                  thickness)
                
                
                buildup = isotope.BuildupHelper.calculate(material,
                                                          gammaray.keV, 
                                                          thickness)
                
                

                dose = dose * attenuation * buildup
                
            
                
                                
            sumdose += dose


        sumdose *= shared.tiac(self.source, self.clearance)
    
            
        sumdose *= self.self_shielding_factor
        
        
        return sumdose
     
    @property
    def self_shielding_factor(self):
        try:
            self_shielding_factor = float(self.source.self_shielding)
        except ValueError:
            self_shielding_factor = 1
        return self_shielding_factor
        
    @property
    def isotope(self):
        if self._isotope is None:

            self._isotope = IsotopeOnGrid(grid=self.dosemap.grid,
                                          position=self.source.position,
                                          isotope_name=self.source.isotope)
        return self._isotope
    
    
    def get_dosemap(self):
        dosemap = np.zeros(self.dosemap.grid.X.shape)
        for keV, gammaray in self.isotope.gammarays.items():
            #time1 = time.time()
            dosemap_keV = gammaray.dosemap
            #time2 = time.time()
            dosemap_keV *= self.attenuation_map(gammaray.keV)
            #time3 = time.time()
            dosemap_keV *= self.buildup_map(gammaray.keV)
            #time4 = time.time()
            dosemap += dosemap_keV
            #time5 = time.time()
            
            #print(time2-time1, time3-time2, time4-time3, time5-time4)
        
        dosemap *= shared.tiac(self.source, self.clearance)
        dosemap *= self.self_shielding_factor
        return dosemap
            
   
    @property
    def material_map(self):
        if self._material_map is None:
            material_map = {}
            for item in self._items:
                material_map = add_dict(material_map, item.material_map)
            
            material_map = add_dict(material_map, self.self_shielding_map)
            
            material_map = {k: v for k,v in material_map.items()\
                            if k != labels.EMPTY_MATERIAL}
            self._material_map = material_map
        return self._material_map
    
    def attenuation_map(self, keV):
        return math.prod([self.material_attenuation_map(material_name, keV)\
                          for material_name in self.material_map.keys()])
            
    def buildup_map(self, keV):
        return math.prod([self.material_buildup_map(material_name, keV)\
                          for material_name in self.material_map.keys()])
            
        
    def material_attenuation_map(self, material_name, keV):
        keV = float(keV)
        if self._attenuation_map is None:
            self._attenuation_map = {}
        
        if (material_name, keV) not in self._attenuation_map.keys():
            attenuation_map = self.get_material_attenuation_map(material_name, keV)
            self._attenuation_map[(material_name, keV)] = attenuation_map            
        return self._attenuation_map[(material_name, keV)]
            
    def get_material_attenuation_map(self, material_name, keV):
        attenuation = None
        
        if material_name in self.material_map.keys():
            material = self.materials.itemByName(material_name)
            thickness_map = self.material_map[material_name]
            
            if material.attenuation_table != labels.EMPTY_TABLE:
                attenuation = isotope.AttenuationHelper.calculate( material, 
                                                                  keV,
                                                                  thickness_map)
            
        if attenuation is None:
            attenuation = np.ones(self.dosemap.grid.X.shape)
            
        return attenuation
    
    def material_buildup_map(self, material_name, keV):
        keV = float(keV)
        if self._buildup_map is None:
            self._buildup_map = {}
        
        if (material_name, keV) not in self._buildup_map.keys():
            buildup_map = self.get_material_buildup_map(material_name, keV)
            self._buildup_map[(material_name, keV)] = buildup_map
        
        return self._buildup_map[(material_name, keV)]
    
    
    def get_material_buildup_map(self, material_name, keV):
  
        thickness = self.material_map[material_name].flatten()
        material = self.materials.itemByName(material_name)
        if any(thickness>0) and material.buildup_table != labels.EMPTY_MATERIAL:

            buildup = isotope.BuildupHelper.calculate(material, keV, thickness)
           
            buildup = buildup.reshape(self.dosemap.grid.X.shape)
            
        else:
            buildup = np.ones(self.dosemap.grid.X.shape)
    
        return buildup

    
    
class Engine:
    _source_dose_map = None
    _sources = None
    _calc_attenuation = None
    
    def __init__(self, dosemap=None, walls=None, shieldings=None,
                 sources=None, clearances=None, materials=None):
       
       
        self.dosemap=dosemap
     
        
        self.walls = walls
        self.shieldings = shieldings
        self.clearances = clearances
        self.materials = materials
    
    @property
    def calc_attenuation(self):
        if self._calc_attenuation is None:
            self._calc_attenuation = isotope.Attenuation()
        return self._calc_attenuation
        
   
        
    def dose_at_point(self, point, sources=None):
  
        self.sources = sources
        
        dose = 0
        
        for item in self._items:
            #calc = lambda: item.dose_at_point(point)
            #print('item', timeit.timeit(calc, number=1))
            dose += item.dose_at_point(point)
        
        
        return dose
    
    @property
    def sources(self):
        if self._sources is None:
            self._sources = []
        return self._sources
    
    @sources.setter
    def sources(self, sources):
        self._items = []
        self._sources = sources
        for source in self.sources:
            if not source.enabled: continue
            clearance = self.clearances.itemByName(source.clearance)
            item = SourceWallsMap(dosemap=self.dosemap,
                                  source=source,
                                  walls=self.walls,
                                  shieldings=self.shieldings,                                 
                                  clearance=clearance,
                                  materials=self.materials,
                                  parent=self)
            
            self._items += [item]
            
    def source_dosemap(self, source):
        self.sources = [source]
        # item = [item for item in self._items if item.source is source][0]
        return self._items[0].get_dosemap()
        
    
    @classmethod
    def from_pyrateshield(cls, model, sources=None):
        #grid = Grid(model.dosemap.extent, model.dosemap.shape)
        return cls(dosemap=model.dosemap, 
                   sources=sources, 
                   walls=model.walls.simplifiedWalls(),
                   shieldings=model.shieldings,
                   clearances=model.clearances,
                   materials=model.materials)
    

        
        
    
    
    
if __name__ == "__main__":
    from pyrateshield.model import Model
    import matplotlib.pyplot as plt
    
    model = Model.load_from_project_file('../../example_projects/LargeProject/project.zip')
    #model = Model.load_from_project_file('../../example_projects/SmallProject/projectTb161.zip')
    #model = Model.load_from_project_file('C:/Users/Marcel/Desktop/test.zip')
   
    
    engine = Engine.from_pyrateshield(model, sources=model.sources_nm)#[model.sources_nm[0]])
    dosemap = lambda: [engine.source_dosemap(source) for source in model.sources_nm]
    #print(timeit.timeit(dosemap, number=1))
    start = time.time()
    for i, point in enumerate(model.critical_points):
        engine.dose_at_point(point.position, sources=model.sources_nm)
        
    stop = time.time()
    
    print(stop-start)
    
       
    

    
    
        
        


# -*- coding: utf-8 -*-
"""
Isotope calculations for pyshield package

Last Updated 05-02-2016
"""
import numpy as np
import scipy.interpolate as si

from pyrateshield.pyshield import tables
from scipy.interpolate import RegularGridInterpolator

class AttenuationHelper:
    _interpolators = {}
    
    @classmethod
    def _get_interpolator(cls, material, energy_keV):
       
        key = (material.attenuation_table, energy_keV)
        
        if key in cls._interpolators.keys():
            return cls._interpolators[key]
        
        table       = tables.ATTENUATION_TABLES[material.attenuation_table]
        
        MeV = np.array(table[tables.ENERGY_MeV])
        mu_p = np.array(table[tables.MASS_ATTENUATION])
        
        interp_fcn = si.interp1d(MeV, mu_p)
  
        cls._interpolators[key] = interp_fcn
        return interp_fcn
    
    @classmethod
    def ulinear(cls, material,  energy_keV):
        umass = cls._get_interpolator(material, energy_keV)(energy_keV/1000) 
        return umass * material.density
    
    @classmethod
    def calculate(cls, material,  energy_keV, thickness):
        return np.exp(-cls.ulinear(material, energy_keV) * thickness)
    
    @classmethod
    def number_mean_free_path(cls, material, energy_keV, thickness):
        return cls.ulinear(material,  energy_keV) *thickness
       
        


class BuildupHelper:
    _interpolators = {}
    _cache = {}
    
    @classmethod
    def _get_interpolator(cls, material, energy_keV):
       
        key = (material.buildup_table, energy_keV)
        
        if key in cls._interpolators.keys():
            return cls._interpolators[key]

        table       = tables.BUILDUP_TABLES[material.buildup_table]
        
        n_mfp       = np.asarray(table[tables.MFP], 'float64')
        
        table       = table.drop(tables.MFP, axis=1) 
        factors     = np.asarray(table, 'float64')
        energies    = np.asarray(table.columns, dtype='float64')
        
        interp_fcn = RegularGridInterpolator((n_mfp,energies), factors,
                                             method='linear', 
                                             bounds_error=False)
        #xi = np.arange(0, 100.1, 0.1)
        xi = n_mfp
        points = [(ii, energy_keV/1000) for ii in xi]
        buildup_keV = interp_fcn(points)
        
        
        interp_fcn = si.interp1d(xi, buildup_keV, kind='linear')
   
        
        # interpolator = si.interp2d(energies, n_mfp, factors)
        # interp_fcn = lambda x, y: cls.interpolant(x, y, interpolator)
        cls._interpolators[key] = interp_fcn
        return interp_fcn
        
        
    @classmethod
    def calculate(cls, material, energy_keV,  thickness):
        
        interpolator = cls._get_interpolator(material, energy_keV)

        n_mfpi = AttenuationHelper.number_mean_free_path(material,energy_keV,  thickness)
        
        if isinstance(n_mfpi, np.ndarray):
            n_mfpi[n_mfpi>100] = 100
        else:
            n_mfpi = min(100, n_mfpi)
        
        value = interpolator(n_mfpi)
        
        return value
    
  


from pyrateshield.floorplan_and_dosemap import MeasuredGeometry, Geometry
from pyrateshield.model_items import  Wall
from pyrateshield.gui.graphics_items import RulerLine
from pyrateshield import labels
from pyrateshield.constants import CONSTANTS
from pyrateshield.gui.item_views import BlockValueChangedSignal



class EditModelItemsController():
    model = None
    _parent = None
    def __init__(self, model=None, view=None, parent=None):
       
        self._parent = parent
        self.view = view
        self.set_model(model)
        
        self.set_callbacks()
        
                
    def parent(self):
        return self._parent

    def set_callbacks(self):
        if self.view.list is not None:
            self.view.list.currentIndexChanged.connect(self.list_selection)
        self.view.valueChanged.connect(self.value_changed)
        if hasattr(self.view, 'new_button'):            
            self.view.new_button.clicked.connect(self.new)
        if hasattr(self.view, 'delete_button'):
            self.view.delete_button.clicked.connect(self.delete)
        self.parent().view.focusSet.connect(self.focus_set)
        
    def focus_set(self, label):
        if label == self.model.item_class.label:
            self.refresh()
    
    def refresh(self):       
        index = self.view.list.currentIndex()
        self.list_selection(index)
            
    
    def value_changed(self, label, value):    
        item = self.item_in_view()
        if item is not None:
            item.update_by_dict({label: value})
    
    def list_selection(self, index):             
        item = self.model.itemAtIndex(index)
        self.item_to_view(item)

    def item_in_view(self):
        return self.model.itemByName(self.view.list.currentText())
        
    def item_to_view(self, item):
        with BlockValueChangedSignal(self.view):
            if item is None:            
                self.view.clear()
                
            elif self.view.list is not None:
                if self.view.list.currentText() != item.name:
                   
                    self.view.list.setCurrentText(item.name)
                   
                else:
                    self.view.setValues(item.to_dict())  
                    
            elif isinstance(item, Wall):
                dct = item.to_dict()
                dct.pop(labels.VERTICES)
                dct.pop(labels.CLOSED)
                self.view.setValues(dct)  
        

        self.view.setEnabled(len(self.model) > 0)

    def delete(self):
        item = self.item_in_view()
        self.model.deleteItem(item)
   
            
    def new(self):

        item = self.item_in_view()
        new_item = self.model.item_class.from_dict(item.to_dict())
        new_item.name = self.model.item_class.default_name
   
        self.model.addItem(new_item)

        #self.view.list.setCurrentText(new_item.name)
       
        
    def disconnect_model(self):        
        if self.model is not None:
            self.model.dataChanged.disconnect()
                     
    def set_model(self, model):  
        self.model = model
        if self.view.list is not None:
            with BlockValueChangedSignal(self.view):
                self.view.list.setModel(self.model)
            
        # if len(self.model) > 0:
        #     self.item_to_view(self.model[0])
            
        self.model.dataChanged.connect(self.item_updated)
        self.model.rowsRemoved.connect(self.item_deleted)
        self.model.rowsInserted.connect(self.item_added)
        

    def item_deleted(self, index, **kwargs):
   
        if len(self.model) == 0:            
            self.view.clear()
            self.view.setEnabled(False)
        else:
            row = index.row()
            if row >= len(self.model):
                row -= 1
            self.item_to_view(self.model.itemAtIndex(row))
    
    def item_added(self, index, **kwargs):
        self.index = index
        self.view.setEnabled(True)
        item = self.model.itemAtIndex(index)
        self.item_to_view(item)
        
        
    def item_updated(self, index, _=None):    
        
        item = self.model.itemAtIndex(index)
 
        if item is self.item_in_view():
            self.item_to_view(item)
       
        
            

class EditSourcesNMController(EditModelItemsController):
    def clearances(self):
        return self.model.parent().clearances
    
    def item_to_view(self, item):
        
        super().item_to_view(item)
        with BlockValueChangedSignal(self.view):
            if self.parent() is None or item is None: return
            self.parent().controllers[labels.CLEARANCE].set_item_by_name(item.clearance)
            
    
    def set_model(self, model):
        super().set_model(model)
        with BlockValueChangedSignal(self.view):
            self.view.clearance_list.setModel(self.clearances())
        
class EditWallsController(EditModelItemsController):
    def disconnect_model(self):
        super().disconnect_model()
        if self.model is not None:
            try:
                self.model.layoutChanged.disconnect()
            except:
                pass
            
    def value_changed(self, label, value):
        super().value_changed(label, value)

            
    def set_callbacks(self):
        super().set_callbacks()
        self.view.scroll_widget.valueChanged.connect(self.sliderMoved)
        self.view.vertex_list.currentIndexChanged.connect(self.vertex_to_view)
        self.view.vertices.valueChanged.connect(self.vertex_to_model)
        
    def sliderMoved(self, index):
        item = self.model.itemAtIndex(index)
        self.item_to_view(item)
        # pixel_item = self.parent().graphics().pixel_item_for_model_item(item)
        # if pixel_item is not None:
        #     pixel_item.marker1.setSelected(True)
        #     pixel_item.marker2.setSelected(True)
            
        
    def refresh(self):
        index = self.view.scroll_widget.value()
        item = self.model.itemAtIndex(index)
        self.item_to_view(item)
    
    def set_model(self, model):        
        super().set_model(model)

        combobox_model = self.shieldings()
        with BlockValueChangedSignal(self.view):
            self.view.shielding_list.setModel(combobox_model)

        self.model.rowsInserted.connect(self.setSliderSize)
        self.model.rowsRemoved.connect(self.setSliderSize)
        self.setSliderSize()
        
    def setSliderSize(self):
        self.view.scroll_widget.setMinimum(0)
        self.view.scroll_widget.setMaximum(self.model.rowCount(0)-1)
    
    def shieldings(self):
        return self.model.parent().shieldings
    
    def item_in_view(self):
        index = self.view.scroll_widget.sliderPosition()
        return self.model.itemAtIndex(index)
    
    def item_to_view(self, item):
        super().item_to_view(item)
        if item is None: return
        
        with BlockValueChangedSignal(self.view):
            if item is not None:
                shielding = self.shieldings().itemByName(item.shielding)
                self.parent().controllers[labels.SHIELDINGS].item_to_view(shielding)
    
            index = self.model.indexForItem(item)
            
            if index is not None:
                self.view.scroll_widget.setSliderPosition(index.row())
            self.update_vertex_list()
    
    def update_vertex_list(self):
     
        item = self.item_in_view()
        current_index = self.view.vertex_list.currentIndex()
        self.view.vertex_list.clear()
        items = [f'Vertex {i}' for i in range(0, len(item.vertices))]
        self.view.vertex_list.addItems(items)
        if len(items) < current_index or current_index is None:
            current_index = 0
        self.view.vertex_list.setCurrentIndex(current_index)
        
    def vertex_to_model(self, value):
      
        item = self.item_in_view()
        index = self.view.vertex_list.currentIndex()
        if index is None or item is None: return
        item.setVertex(index, value)
    
    def marker_to_view(self, marker):
       
        if marker.parentItem() is not self.item_in_view():
            self.item_to_view(marker.parentItem())
        index = marker.parentItem().markers().index(marker)
        self.view.vertex_list.setCurrentIndex(index)
                              
    def set_vertex_index(self, index):
        if index != self.view.vertex_list.currentIndex():
            self.view.vertex_list.setCurrentIndex(index)
            
    
    def vertex_to_view(self):
       
        item = self.item_in_view()
        index = self.view.vertex_list.currentIndex()
        if index is None or index < 0 or item is None: return
        
        vertex = item.vertices[index]
        self.view.vertices.setValue(vertex)
        
        
       
        
        
class EditMaterialsController(EditModelItemsController):
    
    def edit_allowed(self, item):
        return item is not None and item.name != labels.EMPTY_MATERIAL and item.name not in CONSTANTS.base_materials
    
    def disconnect_model(self):
        super().disconnect_model()
        if self.model is not None:
            self.model.nameChanged.disconnect()
            
    def delete_allowed(self, item):
        return item is not None\
            and item not in self.model.items_in_use()\
                and item.name != labels.EMPTY_MATERIAL
    
    def set_model(self, model):        
        super().set_model(model)
        self.model.nameChanged.connect(self.name_changed)
    
    def name_changed(self, old_name, new_name):      
        for shielding in self.model.parent().shieldings:
            if shielding.materials[0][0] == old_name:
                shielding.materials[0][0] = new_name
            elif len(shielding.materials) > 1: 
                if shielding.materials[1][0] == old_name:
                    shielding.materials[1][0] = new_name
                    
                    
    def item_to_view(self, item):
        super().item_to_view(item)
        with BlockValueChangedSignal(self.view):
            self.view.delete_button.setEnabled(self.delete_allowed(item))
            self.view.setEnabled(self.edit_allowed(item))

        
        
class EditClearancesController(EditModelItemsController):
    def disconnect_model(self):
        super().disconnect_model()
        if self.model is not None:
            self.model.nameChanged.disconnect()
            
    def set_model(self, model):
        
        super().set_model(model)
        self.model.nameChanged.connect(self.name_changed)
        
    def name_changed(self, old_name, new_name):
        for source in self.model.parent().sources_nm:
            if source.clearance == old_name:
                source.clearance = new_name
                
    def edit_allowed(self, item):
        return item is not None and item.name != labels.EMPTY_CLEARANCE
                
    def delete_allowed(self, item):
        return item is not None\
            and item not in self.model.items_in_use()\
                and item.name != labels.EMPTY_CLEARANCE
                
    def set_item_by_name(self, name):
        if name in self.model.itemNames():
            self.view.list.setCurrentText(name)
           
                
    def item_to_view(self, item):      
        super().item_to_view(item)      
        with BlockValueChangedSignal(self.view):
            if self.edit_allowed(item):
                self.view.update_available_widgets()
                self.view.delete_button.setEnabled(self.delete_allowed(item))
            else:
                self.view.setEnabled(False)
        
                
class EditShieldingsController(EditModelItemsController):
    
    
        
    def edit_allowed(self, item):
        return item is not None and item.name != labels.EMPTY_SHIELDING 
    
    def disconnect_model(self):
        super().disconnect_model()
        if self.model is not None:
            self.model.nameChanged.disconnect()
    
    def delete_allowed(self, item):
        return item is not None\
            and item not in self.model.items_in_use()\
                and item.name != labels.EMPTY_SHIELDING
                
          
    def set_model(self, model):
        
            
        super().set_model(model)
       
        materials = self.model.parent().materials
        with BlockValueChangedSignal(self.view):
            self.view.materials.material1_list.setModel(materials)
            self.view.materials.material2_list.setModel(materials)
        self.model.nameChanged.connect(self.name_changed)
        
    def name_changed(self, old_name, new_name):        
        for wall in self.model.parent().walls:
            if wall.shielding == old_name:
                wall.shielding = new_name
                
    def item_to_view(self, item):  
        super().item_to_view(item)   
       
        
        self.view.setEnabled(self.edit_allowed(item))
        self.view.delete_button.setEnabled(self.delete_allowed(item))
        
       
            
                


class EditPixelSizeController():
    model = None
    _measured_geometry = None
    _fixed_geometry = None
    _line = None
    
    click_connection = None
    release_connection = None
    move_connection = None
    
    def __init__(self, model=None, view=None, parent=None):
        self.model = model 
        self.view = view
        self.parent = parent
        self.set_callbacks()
        
    def set_callbacks(self):        
        self.view.measure_button.clicked.connect(self.measure_in_view)
        self.view.confirm_button.clicked.connect(self.confirm)
        self.view.valueChanged.connect(self.value_changed)
       
        
        self.parent.view.focusSet.connect(self.focus_set)
        
        
    def line(self):        
        if self._line is None:
         
            self._line = RulerLine(self.geometry(), show_text=False)    
            if self._line.model.distance_pixels == 0:
                self._line.setVisible(False)                              
        return self._line
        
    
    def showLine(self):
        if not isinstance(self.geometry(), MeasuredGeometry):
            self.removeLine()
        else:
            self.graphics().scene().addItem(self.line())                
                                         
    def removeLine(self):
      
        self.graphics().scene().removeItem(self._line)
        self._line = None
        
    def view_geometry_type_changed(self):
        self.showLine()
            
            
    def value_changed(self, label, value):
        if label == labels.GEOMETRY:
            self.view_geometry_type_changed()
        elif label == labels.REAL_WORLD_DISTANCE_CM:
            self.measuredGeometry().distance_cm = value
        elif label == labels.PIXEL_SIZE_CM:            
            self.fixedGeometry().pixel_size_cm = value
        elif label == labels.VERTICES_PIXELS:
            self.measuredGeometry().vertices_pixels = value

    def disconnect_model(self):
        self._measured_geometry = None
        self._fixed_geometry = None
    
    def set_model(self, model):
        self.model = model
        self.item_to_view(self.model)
  

    def confirm(self):        
        self.model.parent().geometry = self.geometry_in_view()        
        self.set_model(self.model.parent().geometry)
        self._line = None
        self.focus_set(labels.PIXEL_SIZE_CM)
    
        
        
    def graphics(self):
        return self.parent.graphics()
    
    def geometry(self):
        if self.view.choose_fixed.isChecked():
            return self.fixedGeometry()
        else:
            return self.measuredGeometry()
                        
    def geometry_in_view(self):
        values = self.view.values()
      
        if self.view.choose_fixed.isChecked():
            geometry = Geometry(pixel_size_cm=values[labels.PIXEL_SIZE_CM],
                                origin=labels.TOP_LEFT)
        elif self.view.choose_measured.isChecked():
            if values[labels.REAL_WORLD_DISTANCE_CM] == 0:
                values[labels.REAL_WORLD_DISTANCE_CM] = 1
            geometry = MeasuredGeometry(vertices_pixels=values[labels.VERTICES_PIXELS],
                                        distance_cm=values[labels.REAL_WORLD_DISTANCE_CM],
                                        origin=labels.TOP_LEFT)
            
        # geometry.setParent(self.model.parent())
        return geometry
    
    def geometry_updated(self, event_data):       
        src, label, old_value, value = event_data
                
        self.view.setValue(label, value)
        
        if isinstance(self.geometry(), MeasuredGeometry):
            self.view.setValue(labels.PIXEL_SIZE_CM, self.geometry().pixel_size_cm)
        

    def item_to_view(self, item):
        with BlockValueChangedSignal(self.view):
            if isinstance(item, MeasuredGeometry):
                self.view.set_choose_measured()
            else:
                self.view.set_choose_fixed()
            
            if item is not None:
                dct = item.to_dict()
                dct.pop(labels.ORIGIN, None)
                self.view.setValues(dct)
           

    def measuredGeometry(self):
        if self._measured_geometry is None:
            if isinstance(self.model, MeasuredGeometry):
                geometry = MeasuredGeometry.from_dict(self.model.to_dict())
                self._measured_geometry = geometry
            else:
                self._measured_geometry = MeasuredGeometry(origin=labels.TOP_LEFT)
            self._measured_geometry.update_event.connect(self.geometry_updated)
        return self._measured_geometry
    
    def fixedGeometry(self):
        if self._fixed_geometry is None:
            if not isinstance(self.model, MeasuredGeometry):
                geometry = Geometry.from_dict(self.model.to_dict())
                self._fixed_geometry = geometry
            else:
                self._fixed_geometry = Geometry(origin=labels.TOP_LEFT)
            self._fixed_geometry.update_event.connect(self.geometry_updated)
        return self._fixed_geometry
    
    
    def focus_set(self, label):    
        if label == labels.PIXEL_SIZE_CM:
            self.graphics().scene().setVisibleGraphicsModelItems(False)
            self.item_to_view(self.model)
            self.view_geometry_type_changed()
            self.parent.view.toolbar.setEnabled(False)
            self.parent.view.result_container.setTabEnabled(1, False)
            self.showLine()
        else:
            self.graphics().scene().setVisibleGraphicsModelItems(True)
            
            self.parent.view.toolbar.setEnabled(True)
            self.parent.view.result_container.setTabEnabled(1, True)
            self._measured_geometry = None
            self._fixed_geometry = None
            self.removeLine()
            
        
              
    def measure_in_view(self):    

        def start(event):
           
            
            pos = self.graphics().mapToScene(event.pos())
            vertices_pixels = [[pos.x(), pos.y()], [pos.x(), pos.y()]]           
            self.line().model.vertices_pixels = vertices_pixels
            self.move_connection = self.graphics().mouseMove.connect(update)
            self.line().setVisible(True)

        def update(event):
            #print('update')
            pos = self.graphics().mapToScene(event.pos())                     
            self._line.model.set_vertex_pixels(1, [pos.x(), pos.y()])
         
        def disconnect():
            if self.click_connection is not None:
                self.graphics().leftMouseClick.disconnect(self.click_connection)
            if self.release_connection is not None:
                self.graphics().mouseRelease.disconnect(self.release_connection)
            if self.move_connection is not None:            
                self.graphics().mouseMove.disconnect(self.move_connection)
            
            self.click_connection = None
            self.release_connection = None
            self.move_connection = None
        
        def end(event):
            update(event)
            disconnect()
            
                        
        self.click_connection = self.graphics().leftMouseClick.connect(start)        
        self.release_connection = self.graphics().mouseRelease.connect(end)
        
        
            
        
        
        
        
    

        



from PyQt5 import QtGui
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import Qt, QObject, QEvent

from pyrateshield.model import Model
from pyrateshield.dosemapper import Dosemapper
from pyrateshield.model_items import Wall, ModelItem, Ruler

from pyrateshield.gui import io, calculator, dosemap_style, isotopes
from pyrateshield.gui.main_view import MainView, TOOLBOX_LAYOUT
from pyrateshield.gui.graphics_items import  GraphicsModelItem, WallMarker
from pyrateshield.gui.toolbar import ToolbarController
from pyrateshield.gui.context_menu import ContextMenuController
from pyrateshield.gui.critical_point_controller import CriticalPointReportController
from pyrateshield.gui.toolbar import Toolbar
from pyrateshield.gui.item_controllers import (EditModelItemsController, 
                                               EditSourcesNMController,                                               
                                               EditWallsController, 
                                               EditClearancesController, 
                                               EditShieldingsController,
                                               EditPixelSizeController,
                                               EditMaterialsController)
                                               
                                                                                              
from pyrateshield import labels


class MainController():
    _mpl_controller = None
    _critical_points_controller = None
    _calculator = None
    _dosemap_style = None
    _old_status2_msg = None
    model = None
    
    
    def __init__(self, dosemapper=None, model=None, view=None):
        
        if dosemapper is None:
            dosemapper = Dosemapper(multi_cpu=False)
        
        if model is None:
            model = self.get_empty_model()
    
        if view is None:
            view = MainView()
        
        self.last_selected_item = None
        
        self.dosemapper = dosemapper
        self.view = view
        self.model=model
        
        self.controllers = self.create_controllers()

        self.connect_model()
        
        self.set_view_callbacks()

        self.controllers[labels.SHIELDINGS].view.setEnabled(True)
        
        self.graphics().scene().updateStyles()
        
        self.graphics().mouseMove.connect(self.updateMousePosition)
        
        self.model.history.reset()
        
        self.dosemapper.progress.connect(self.set_progress)
        self.dosemapper.calculation_time.connect(self.calculation_time)
        self.dosemapper.interpolate.connect(self.interpolate_msg)
        
    def set_progress(self, value):
        value = int(round(value*100))
        value = max(0, value)
        value = min(value, 100)
        self.view.progress.setValue(value)
        self.view.progress.update()
        
    def interpolate_msg(self, interpolating):
        if self._old_status2_msg is None:
            self._old_status2_msg = self.view.status_label2.text()
        if interpolating == -1:
            self.view.status_label2.setText('Interpolating.....')
        else:
            time = round(interpolating, 1)
            if time == 0:
                time = '< 0.1'
            new_msg = self._old_status2_msg
            new_msg += f' | Interpolation time: {time} s'
            self.view.status_label2.setText(new_msg)
            self._old_status2_msg = None
        
    def calculation_time(self, time):
        time = round(time, 1)
        if time == 0:
            time = '< 0.1'
        self.view.status_label2.setText(f'Calculation time: {time} s')
        
    def updateMousePosition(self, event):
        pos = self.graphics().mapToScene(event.pos())
        psize = self.model.get_pixel_size_cm()
        
        x = round(pos.x() * psize, 1)
        y = round(pos.y() * psize, 1)
        
        text = f'Mouse position [cm]: {x}, {y}'
        
        self.view.status_text = text
        

    def graphics(self):
        return self.view.views[labels.CANVAS]
        
    def disconnect_model(self):
        for key, controller in self.controllers.items():           
            controller.disconnect_model()
            
    def set_model(self, model):     
        self.view.clear()              
        self.disconnect_model()
        self.model = model
        self.connect_model()
        
    def connect_model(self):           
        self.graphics().setModel(self.model)
        self.toolbar_controller.model = self.model
        for key, controller in self.controllers.items():
            model = self.model.get_attr_from_label(key)
            controller.set_model(model)
   
        
        self.context_menu_controller.set_model(self.model)
        self.crit_point_controller.set_model(self.model)
        item = None
        if len(self.model.sources_nm) > 0:
            item = self.model.sources_nm[0]
        elif len(self.model.sources_ct) > 0:
            item = self.model.sources_ct[0]
        elif len(self.model.sources_xray) > 0:
            item = self.model.sources_xray[0]
        elif len(self.model.critical_points) > 0:
            item = self.model.critical_points[0]
        elif len(self.model.walls) > 0:
            item = self.model.walls[0]
        if item is not None:
            self.select_item(item)
            
        
    def itemChanged(self, *args, **kwargs):
        pass
   

    def set_view_callbacks(self):
        for label, tab in self.view.toolbox_tabs.items():            
            if len(TOOLBOX_LAYOUT[label]) > 1:                
                tab.currentChanged.connect(self.tab_selected)
                
        self.view.toolbox.currentChanged.connect(self.toolbox_selected)
       
       
        self.view.toolbar.buttonClick.connect(self.toolbar_callback)
        
        #self.graphics().scene().selectionChanged.connect(self.graphics_selection_changed)
        
        self.graphics().leftMouseClick.connect(self.select_by_mouse)
        
        
    def select_by_mouse(self, _, item):
        self.select_item(item)

    def select_item(self, item):   
        # if isinstance(item, Marker):
        #     item = item.parentItem()
            
        if isinstance(item, GraphicsModelItem):
            item = item.model

        if isinstance(item, Wall):
            self.view.set_focus(labels.WALLS)                            
            self.controllers[labels.WALLS].item_to_view(item)
            
        elif isinstance(item, WallMarker):
            controller = self.controllers[labels.WALLS]
            wall = item.parentItem().model
            
            if wall is not controller.item_in_view():
                controller.item_to_view(wall)
            
        
            index = item.parentItem().markers().index(item)
            
            controller.set_vertex_index(index)
            
            
            item.scene().silentSelectItem(item)
            
            
            self.view.set_focus(labels.WALLS)             
            
        elif isinstance(item, ModelItem) and not isinstance(item, Ruler):        
            self.view.set_focus(item.label)
            self.controllers[item.label].item_to_view(item)            
        

      
    @property
    def controller_types(self):
        return {
                labels.SOURCES_CT:        EditModelItemsController,
                labels.SOURCES_XRAY:      EditModelItemsController,
                labels.SOURCES_NM:        EditSourcesNMController,
                labels.CLEARANCE:         EditClearancesController,
                labels.WALLS:             EditWallsController,
                labels.SHIELDINGS:        EditShieldingsController,
                labels.CRITICAL_POINTS:   EditModelItemsController,
                labels.PIXEL_SIZE_CM:     EditPixelSizeController,
                labels.MATERIALS:         EditMaterialsController}       

        
    def create_controllers(self):
        # FIX!: dosemapper=self.dosemapper,
        
        view = self.view.views[labels.CANVAS]
        
    
        
        self.toolbar_controller = ToolbarController(view=self.view.toolbar,
                                                    model=self.model,
                                                    parent=self)
                                                  
        
        self.context_menu_controller = ContextMenuController( 
            model=self.model, view=self.graphics(), parent=self)
        
        view = self.view.views[labels.CRITICAL_POINT_REPORT_VIEW]
        self.crit_point_controller = CriticalPointReportController(
            view=view, model=self.model, 
            controller=self, dosemapper=self.dosemapper)
        
    
        controllers = {}
        for key in self.controller_types.keys():
                    # labels.CRITICAL_POINTS, labels.PIXEL_SIZE_CM, 
                    #  labels.MATERIALS):
           
            contr_type = self.controller_types[key]
            view = self.view.views[key]
            model = self.model.get_attr_from_label(key)
            controllers[key] = contr_type(model = model, 
                                          view = view,
                                          parent = self)
            
        
                                         
        return controllers

            
    def toolbox_selected(self, _=None):
        self.tab_selected()
        
        
    def tab_selected(self, index=None):        
        label = self.view.get_active_tool_panel_name()
        self.view.focusSet.emit(label)
        
    def toolbar_callback(self, toolname):
        if toolname == Toolbar.SAVE_AS:
            self.save_as()
        elif toolname == Toolbar.LOAD:
            self.load()
        elif toolname == Toolbar.SAVE:
            self.save()
        elif toolname == Toolbar.IMAGE:
            self.load_image()
        elif toolname == Toolbar.NEW:
            self.new()
        elif toolname == Toolbar.CALCULATOR:
            self.show_calculator()
        elif toolname == labels.DOSEMAP_STYLE:
            self.show_dosemap_style()
        elif toolname == Toolbar.SNAPSHOT:
            self.snapshot()
        elif toolname == Toolbar.ISOTOPES:
            self.show_isotopes()
           
    def snapshot(self):
        file = io.ask_new_image_file('Save snapshot to disk')
        if file is not None:
          
            
            viewport = self.view.views[labels.CANVAS].viewport()
            pixmap = QtGui.QPixmap(viewport.size())
            painter = QtGui.QPainter(pixmap)
            viewport.render(painter)
            painter.end()
            
            pixmap.save(file)
            
            
    def load_image(self):
       image = io.safe_load_image()
       if image is not None:
           
           # Load floorplan and reset pixel size (geometry)
           self.model.floorplan.image = image                                                    
           self.view.set_focus(labels.PIXEL_SIZE_CM)
           
    def new(self):        
        confirm = io.confirm_new_project()
        if confirm:
            self.set_model(self.get_empty_model())
        self.view.set_focus(labels.SOURCES_NM)
        
            
    def save(self):
        io.safe_write_model(model=self.model, 
                                        filename=self.model.filename)
            
    def save_as(self):
        io.safe_write_model(model=self.model)
               
    def load(self):
        model = io.safe_load_model(self.model.filename)        
        if model is not None:
            self.set_model(model)
               
    def show_calculator(self):
        self._calculator = calculator.Controller()
        self._calculator.view.show()
    
    def show_dosemap_style(self):
        self._dosemap_style = dosemap_style.Controller(self.model,
                                                       parent=self)
        self._dosemap_style.view.show() 
        
    def show_isotopes(self):
        self._isotopes = isotopes.Controller()
        self._isotopes.view.show()
        self._isotopes.view.resize(500, 300)
        
        
    def get_empty_model(self):
        model = Model()
        model.floorplan.geometry.origin = labels.TOP_LEFT
        return model

def main(model=None, controller=None, dm=None):
        app = QApplication([])
        
        controller = MainController(model=model, dosemapper=dm)
        window = controller.view
        window.show()    
        app.exec_()
      
if __name__ == "__main__":  
    project = '../../example_projects/SmallProject/project.zip'
    project = '../../example_projects/LargeProject/project.zip'
    #project = '/Users/marcel/Downloads/T1 NG_20230828_Self-shielding pt 20cm water_All Sources On_v12.zip'
    #model = Model.load_from_project_file(project)
    #file = 'C:/Users/r757021/Desktop/test.zip'
    model = Model.load_from_project_file(project)
    #model = Model()
    
    app = QApplication([])
    dosemapper = Dosemapper(multi_cpu=True)
    #controller = MainController(model=Model(), dosemapper=None)
    controller = MainController(model=model, dosemapper=dosemapper)

    window = controller.view
    window.show()    
    
    # global_event_filter = GlobalEventFilter(model=controller.model)
    # app.installEventFilter(global_event_filter)
    
    # model = Model.load_from_project_file(project)
    app.exec_()
    
    



import os
import zipfile
import io
import yaml
import pickle
import imageio


from pyrateshield.constants import CONSTANTS
from pyrateshield.modelitem import ModelItem

from pyrateshield.model_sequences import (Walls,
                                          SourcesNM,
                                          SourcesCT,
                                          SourcesXray,
                                          CriticalPoints,
                                          Shieldings,
                                          Materials,
                                          Clearances,
                                          Rulers)
                                        
from functools import partial

from pyrateshield.floorplan_and_dosemap import (Dosemap, Floorplan, 
                                                DosemapStyle, MeasuredGeometry)
from pyrateshield import labels

try:
    wdir = os.path.split(__file__)[0] 
except:
    wdir = os.getcwd()
    
filename = os.path.join(wdir, 'defaults.yml')
    
DEFAULTS = yaml.safe_load(open(filename))


class PauseHistoryRecording:
    def __init__(self, history):
        self.history = history
    def __enter__(self):
        self.history.block_recording(True)
    def __exit__(self, *args):
        self.history.block_recording(False)

class PauseHistoryCleanup:
    def __init__(self, history):
        self.history = history
    def __enter__(self):
        self.history.block_cleanup(True)
    def __exit__(self, *args):
        self.history.block_cleanup(False)


class ModelHistory:
    ATTRIBUTE_VALUE_CHANGE      = 'attribute_value_change'
    MODEL_ITEM_DELETED          = 'model_item_deleted'
    MODEL_ITEM_ADDED            = 'model_item_added'
    
    _block_recording = False
    _undo_records = None
    _stack = None
    _block_cleanup = False
    
    def __init__(self, parent=None):
        self.parent = parent
        self.set_callbacks()
        
    def stack(self):
        if self._stack is None:
            self._stack = {}
        return self._stack
    
    def undo_records(self):
        if self._undo_records is None:
            self._undo_records = []
        return self._undo_records
    
    def redo_records(self):
        if self._redo_records is None:
            self._redo_records = []
        return self._redo_records
        
    def add_to_stack(self, record):
        # stack can only collect a list of change that are the same type
        type_of_change, _ = record
        if type_of_change not in self.stack().keys():
            if len(self.stack()) > 0:
                self.clean(force=True)
            self.stack()[type_of_change] = []
        self.stack()[type_of_change].append(record)
      

    def set_callbacks(self):
        # listen to changes in the model
        for attr in self.parent._attr_dct.keys():
            item = getattr(self.parent, attr)
            cbk = partial(self.add_undo_record, self.ATTRIBUTE_VALUE_CHANGE, item)
            if isinstance(item, ModelItem):
                # Non model sequences, like floorplan and dosemap
                pass
            else:
                # model sequences
                item.valueChanged.connect(cbk)
                cbk = partial(self.add_undo_record, self.MODEL_ITEM_DELETED, item)
                item.rowsAboutToBeRemoved.connect(cbk)
                cbk = partial(self.add_undo_record, self.MODEL_ITEM_ADDED, item)
                item.rowsInserted.connect(cbk)
                
    def add_undo_record(self, type_of_change, sequence, event_data):
         if self.isBlockedRecording(): return
         self._redo_records = None
         if type_of_change == self.ATTRIBUTE_VALUE_CHANGE:  
             item, label, old_value, new_value = event_data
             if label == labels.CLOSED:
                 self.clean()
             elif label in (labels.VERTICES, labels.VERTICES_PIXELS):
                 
                 if old_value == new_value:
                     return
                 if len(old_value) != len(new_value):
                     self.clean()

             self.add_to_stack([type_of_change, event_data])
                 
         elif type_of_change in (self.MODEL_ITEM_DELETED,
                                 self.MODEL_ITEM_ADDED):
             model_item = sequence.itemAtIndex(event_data)
             self.add_to_stack([type_of_change, (sequence, model_item)])   
             
    def items_to_clean(self):
        if len(self.stack()) == 0: 
            return []
        else:
            key = list(self.stack().keys())[0]
            return self.stack()[key]
        
    def clean(self, force=False):    
        # Tricky implementation. All model changes are recorded in stack.
        # When the selection in the UI changes or undo is called all entries
        # stored in stack and stored in records. Except for position changes.
        # For position changes only the first record is stored to records. So
        # when items are moved by the mouse the original location is kept and 
        # all other locations are ignored. Undo will then restore the location
        # where the movement initiated.
        
        if self._block_cleanup and not force: return
      
        items_with_position_changes = {}
        records = []
        pos_labels = (labels.VERTICES_PIXELS, labels.POSITION, labels.VERTICES)
        
        for record in self.items_to_clean(): 
            type_of_change, data = record
            data = list(data)
            record = [type_of_change, data]
            if type_of_change != self.ATTRIBUTE_VALUE_CHANGE:
                records.append(record)
            else:
                item, label, old_value, new_value = data
                if label not in pos_labels:
                    records.append(record)
                else:
                    if item not in items_with_position_changes.keys():
                        items_with_position_changes[item] = record
                        records.append(record)
                    else:
                        _, data = items_with_position_changes[item]
                        data[-1] = new_value
                        if len(data) != 4:
                            raise

        if len(records) > 0:
            self.undo_records().append(records)
        
        self._stack = None
        
    def block_cleanup(self, block):
        self._block_cleanup = block

    def get_entries_to_undo(self):
        self.clean(force=True)
        return self.undo_records().pop(-1) if len(self.undo_records()) > 0 else []
    
    def get_entries_to_redo(self):
        return self.redo_records().pop(-1) if len(self.redo_records()) > 0 else []
                     
    def block_recording(self, block):
        self._block_recording = block
        
    def isBlockedRecording(self):
        return self._block_recording
                
    def undo(self):
        undo_items = self.get_entries_to_undo()
        for entry in undo_items:
            self.undo_entry(entry)
        self.redo_records().append(undo_items)
            
    
    def redo(self):
        
        redo_items = self.get_entries_to_redo()
     
        for entry in redo_items:
            self.redo_entry(entry)
        self.undo_records().append(redo_items)
        
    def redo_entry(self, record):
        type_of_change, data = record
       
        if type_of_change == self.ATTRIBUTE_VALUE_CHANGE:
            item, label, _, new_value = data
            attr = item.attr_name_from_label(label)
            with PauseHistoryRecording(self):
                setattr(item, attr, new_value)
        elif type_of_change == self.MODEL_ITEM_DELETED:     
            sequence, item = data
            with PauseHistoryRecording(self):
                self.parent.delete_item(item)
        elif type_of_change == self.MODEL_ITEM_ADDED:
            sequence, item = data
            with PauseHistoryRecording(self):
                sequence.addItem(item)
        
                
        
        
    def undo_entry(self, record):
        type_of_change, data = record
       
        if type_of_change == self.ATTRIBUTE_VALUE_CHANGE:
            item, label, old_value, _ = data
            attr = item.attr_name_from_label(label)
            with PauseHistoryRecording(self):
                setattr(item, attr, old_value)
        elif type_of_change == self.MODEL_ITEM_DELETED:     
            sequence, item = data
            with PauseHistoryRecording(self):
                sequence.addItem(item)
        elif type_of_change == self.MODEL_ITEM_ADDED:
            _, item = data
            with PauseHistoryRecording(self):
                self.parent.delete_item(item)
                
        
                
    def reset(self):      
        self._records = None
        self._stack = None
    
    
            

class Model(ModelItem):
    
    filename            = None

    _attr_defaults = {labels.SOURCES_NM:        SourcesNM,
                      labels.SOURCES_CT:        SourcesCT,
                      labels.SOURCES_XRAY:      SourcesXray,
                      labels.CRITICAL_POINTS:   CriticalPoints,
                      labels.WALLS:             Walls,
                      labels.CLEARANCE:         Clearances,
                      labels.MATERIALS:         Materials,
                      labels.SHIELDINGS:        Shieldings,
                      labels.FLOORPLAN:         Floorplan,
                      labels.DOSEMAP:           Dosemap,
                      labels.DOSEMAP_STYLE:     DosemapStyle,
                      labels.RULERS:            Rulers}
    
    _attr_dct = {
        "sources_nm":       labels.SOURCES_NM,
        "sources_ct":       labels.SOURCES_CT,
        "sources_xray":     labels.SOURCES_XRAY,
        "critical_points":  labels.CRITICAL_POINTS,
        "clearances":       labels.CLEARANCE,
        "materials":        labels.MATERIALS,
        "shieldings":       labels.SHIELDINGS,
        "floorplan":        labels.FLOORPLAN,
        "dosemap":          labels.DOSEMAP,
        "dosemap_style":    labels.DOSEMAP_STYLE,
        "walls":            labels.WALLS,
        "rulers":           labels.RULERS
        }
    
    
   
   
    
    
    def __init__(self, **kwargs):
        self.constants = CONSTANTS
        
        kwargs = self.prepare_init(**kwargs)
        
  
        ModelItem.__init__(self, **kwargs)
        
        self.set_defaults()
        
        # for attr in self._attr_dct.keys():
        #     getattr(self, attr).setParent(self)
        
        

        if self.floorplan.geometry.origin == labels.BOTTOM_LEFT:
            # PyQt5 has top left as (0, 0), matplotlib bottom left
            # Flip all items in y if model is saved with matplotlib
            self.set_legacy_y()

        self.fix_none_entries()
        self.history = ModelHistory(parent=self)
    
    # Code to make objects pickable
    def __getstate__(self):
        dct = self.to_dict()
        # floorplan not in dict (!)
        dct[labels.FLOORPLAN][labels.IMAGE] = self.floorplan.image
        return dct
        
    def __setstate__(self, dct):
        self.__init__(**dct)
        # restore floorplan
        self.floorplan.image = dct[labels.FLOORPLAN][labels.IMAGE]
    # end pickable
    
    
    def set_defaults(self):
        if len(self.clearances) == 1:
            self.clearances = Clearances.from_list(DEFAULTS[labels.CLEARANCE])
        if len(self.materials) == 1:           
            self.materials = Materials.from_list(DEFAULTS[labels.MATERIALS])
        if len(self.shieldings) == 1:
            self.shieldings = Shieldings.from_list(DEFAULTS[labels.SHIELDINGS])
    
    def get_attr_from_label(self, label):
        if label in (labels.PIXEL_SIZE_CM, labels.GEOMETRY):
            return self.floorplan.geometry
        else:
            return super().get_attr_from_label(label)
                
    def prepare_init(self, **kwargs):
        for key, value in kwargs.items():
            if key in self._attr_dct.keys():
                label = self._attr_dct[key]
            else:
                label = key
            if isinstance(value, list):
                kwargs[key] = self._attr_defaults[label].from_list(value)
            elif isinstance(value, dict):
                kwargs[key] = self._attr_defaults[label].from_dict(value)
        return kwargs
        
    def to_dict(self):
        dct = {}
        for attr in self._attr_dct.keys():
            value  = getattr(self, attr)
            if hasattr(value, 'to_dict'):
                value = value.to_dict()
            elif hasattr(value, 'to_list'):
                value = value.to_list()
            else:
                raise AttributeError('Cannot serialize {attr}')
            dct[self._attr_dct[attr]] = value
        return dct
    
    def __str__(self):
        return yaml.dump(self.to_dict())

    def delete_item(self, item):
        sequence = self.sequence_for_item(item)
        if item in sequence:
            sequence.deleteItem(item)
    
    def delete_items(self, items):
        with PauseHistoryCleanup(self.history):
            for item in items:
                self.delete_item(item)
        self.history.clean()

    def add_item(self, item):
        # add a item of any type used by gui
        sequence = self.sequence_for_item(item)
        sequence.addItem(item)
 
    def sequence_for_item(self, item):
        return self.get_attr_from_label(item.label)
    
    def sort_model_items(self, items):
        # Used by UI to sort a bunch of selected items
        sorted_items = {}
        
        for item in items:
            if item.label not in sorted_items.keys():
                sorted_items[item.label] = []
            sorted_items[item.label] += [item]        
        return sorted_items
    
    def item_class_for_label(self, label):
        sequences = {label: item for label, item in self._attr_defaults.items()\
                     if hasattr(item, 'item_class')}
       
        
        if label in sequences.keys():
            return sequences[label].item_class
        
        else:
            raise KeyError(f'No sequence defined for label {label}')

    def save_to_project_file(self, filename):    
        file, ext = os.path.splitext(filename)
        if ext.lower() != '.zip':
            filename = file + '.zip' # force saving old projects to ext zip
        
        self.filename = filename
                        
        project_dict = self.to_dict()
        image = self.floorplan.image
        
        temp_yml = io.StringIO()
        yaml.dump(project_dict, temp_yml, default_flow_style=None)
        
        temp_img = io.BytesIO()
        imageio.imwrite(temp_img, image, format=".png")

        with zipfile.ZipFile(filename, "w") as zf:
            zf.writestr("project.yml", temp_yml.getvalue())
            zf.writestr("floorplan.png", temp_img.getvalue())
    
    
    @classmethod
    def load_from_project_file(cls, filename):
        try:
            with zipfile.ZipFile(filename) as zf:
                with zf.open("floorplan.png", "r") as f:
                    image = imageio.imread(f)
                with zf.open("project.yml", "r") as f:
                    dct = yaml.safe_load(f)
                 
                
                dct[labels.FLOORPLAN][labels.IMAGE] = image
                model = cls.from_dict(dct)
                
        
        except zipfile.BadZipFile:
            ### For backwards compatibility:
            success = False
            if filename.endswith(".psp"):
                try:
                    model = cls.load_old_psp_file(filename)
                    success = True
                except:
                    pass
                
            if not success:
                # IOError is picked up by GUI to show error dlg
                raise IOError(f'Could not read {filename}')
    
        model.filename = filename
        return model
    
    
# =============================================================================
#     LEGACY STUFF
# =============================================================================
    
    @classmethod
    def load_old_psp_file(cls, filename):
        ### For backwards compatibility:
        with open(filename, 'rb') as fp:
            dct = pickle.load(fp)
        
        image = dct.pop("IMAGE_DATA")   
        dct[labels.FLOORPLAN][labels.IMAGE] = image
        dct[labels.FLOORPLAN].pop('Filename', None)
        
        # ugly code to remove invalid key from older versions
        dct[labels.FLOORPLAN][labels.GEOMETRY].pop(False, False)
        
        return cls.from_dict(dct)

    # def zero_origin(self):
    #      # old psp files had an origin option, origin is now always at (0, 0)
         
    #      origin = self.floorplan.geometry.origin_cm
         
    #      if origin[0] != 0 or origin[1] != 0:
    #          self.shift_cm(origin[0], origin[1])
    #          self.floorplan.geometry.origin_cm = [0, 0]  
    
    
    def fix_none_entries(self):
        for shielding in self.shieldings:
            materials = shielding.materials
            if materials[0][0] in (None, ''):
                materials[0][0] = Materials.empty_item_name
            if len(materials) > 1:
                if materials[1][0] in (None, ''):
                    materials[1][0] = Materials.empty_item_name
        
        for wall in self.walls:
            if wall.shielding in (None, ''):
                wall.shielding = Shieldings.empty_item_name
                
        # for sources in self.sources_nm:
        #     if source.
        
    
    def set_legacy_y(self):
        print('SET LEGACY Y COORDS')
        psize = self.get_pixel_size_cm()
        height = self.floorplan.image.shape[0]  
        for wall in self.walls:
            
            wall.vertices[0][1] = height* psize - wall.vertices[0][1]
            wall.vertices[1][1] = height* psize - wall.vertices[1][1]
        
        
        for point in [*self.sources_ct,
                      *self.sources_xray,
                      *self.sources_nm,
                      *self.critical_points]:
           
            point.position[1] = height* psize - point.position[1]
        
        
        
        if isinstance(self.floorplan.geometry, MeasuredGeometry):
            vvp = self.floorplan.geometry.vertices_pixels
            vvp[0][1] = height - vvp[0][1]
            vvp[1][1] = height - vvp[1][1]
        
        self.floorplan.geometry.origin = labels.TOP_LEFT

if __name__ == "__main__":
    # import yaml
    # with open('test_model_in.yml') as file:
    #     dct = yaml.safe_load(file)
    # model = Model.from_dict(dct)
    file = '../example_projects/SmallProject/project.zip'
    file = '../example_projects/LargeProject/project.zip'
    # file = 'C:/Users/757021/git/pyrateshield/example_projects/Nucleaire Geneeskunde.psp'
    #file = 'C:/Users/r757021/Desktop/test.zip'
    model = Model.load_from_project_file(file)
    # model = Model()
    



NAME = "Name"
KVP = "kVp"

### Constants file
SHOW_LEGEND = 'Show legend'

CRITICAL_POINT_DOSE_CORRECTED = 'Dose corrected for occupancy [mSv]'
CRITICAL_POINT_DOSE = 'Dose [mSv]'

RULERS = 'Rulers'
                             
WALLS_AND_SHIELDING = "Walls && Shieldings"
SOURCES_NM_CT_XRAY = "Sources NM, CT, XRAY"
PROJECT = 'Project'


PYSHIELD = 'PyShield'
RADTRACER = 'Radtracer'


LOAD_SAVE = "Load/Save"
PREFERENCES = "Preferences"
CRITICAL_POINT_REPORT_VIEW = "Critical Point Report"
CANVAS = "Canvas"

ENABLED = 'Enabled'


# Materials
BASE_MATERIALS = 'Base Materials'
BUILDUP_MATERIALS = 'Buildup Materials'
MATERIALS = "Materials"
MATERIAL = 'Material'
DENSITY = "Density"
THICKNESS = "Thickness [cm]"
BUILDUP_TABLE = 'Buildup Table'
ATTENUATION_TABLE = 'Attenuation Table'
RADTRACER_MATERIAL = 'Radtracer Material'
EMPTY_MATERIAL = 'None'
EMPTY_SHIELDING = 'None'
EMPTY_TABLE = 'None'

# Isotopes
ISOTOPES = "Isotopes"
HALF_LIFE = "Half life [h]"
SELF_SHIELDING_OPTIONS = "Self shielding options"
SELF_SHIELDING_NONE = 'None'
SELF_SHIELDING_BODY = 'Body'
SELF_SHIELDING_FACTOR = 'Factor'
DECAY_CHAINS = "Decay chains"
ISOTOPE_SPECTRA = "Isotope spectra"
SPECTRUM = "Spectrum"
SUPPORTED_ISOTOPES = 'Supported isotopes'


# Archer parameters
CT_BODY_PART_OPTIONS = "CT Body part options"
CT_PARAMETERS = "CT parameters"
XRAY_PARAMETERS = "Xray parameters"
ARCHER = "Archer"
CT_SCATTER_FRACTION_BODY = "Scatter fraction body"
CT_SCATTER_FRACTION_HEAD = "Scatter fraction head"
XRAY_SCATTER_FRACTION = "Scatter fraction"
# 


### Model file
ENGINE = "Engine"
RADTRACER = "Radtracer"
PYSHIELD = "pyShield"

# Floor plan
FLOORPLAN = "Floor plan"
IMAGE = 'image'
FILENAME = "Filename"
PIXEL_SIZE_CM = "Pixel size [cm]"
ORIGIN_CM = "Origin [cm]"
ORIGIN = 'Origin'
TOP_LEFT = "top_left"
BOTTOM_LEFT = "bottom_left"
PIXEL_SIZE_METHOD = 'Pixel size method'
FIXED = 'fixed' # pixel size set by hand
MEASURED = 'measured' # measured form floor plan
REAL_WORLD_DISTANCE_CM = 'Real world distance [cm]' # distance set in gui
VERTICES_PIXELS = 'vertices [pixels]'

# Plot style
DOSEMAP_STYLE = "Dose map style"
CMAP_NAME = "Colormap name"
CMAP_MIN = "Colormap min"
CMAP_MAX = "Colormap max"
CMAP_ALPHA = "Transparency"
CMAP_ALPHA_GRADIENT = "Transparency gradient"
CONTOUR_LINES = "Contour lines"

# Dose map
RESULT = 'Result'
DOSEMAP = "Dose map"
GEOMETRY = "Geometry"
GRID_MATRIX_SIZE = "Grid matrix size"
EXTENT = "Extent [cm]"
INTERPOLATE = 'Interpolate'
MULTI_CPU = 'Multi CPU'

# Model
CRITICAL_POINTS = "Critical points"
SHIELDINGS = "Shieldings"
WALLS = "Walls"
SOURCES_CT = 'Sources CT'
SOURCES_NM = 'Sources NM'
SOURCES_XRAY = 'Sources Xray'


### General
POSITION = "Position"

SELECTED_ENGINE = 'Selected Engine'
    
### Shielding
COLOR = "Color"
LINEWIDTH = "Linewidth [pt]"
MATERIAL_NAME = "Material name"
MATERIAL_THICKNESS = "Thickness [cm]"

### Wall
SHIELDING = "Shielding"
VERTICES = "Vertices"
CLOSED = "Closed"

### Critical point
OCCUPANCY_FACTOR = "Occupancy factor"

### Sources    
NUMBER_OF_EXAMS = "Number of exams"

# Nuclear
ISOTOPE = "Isotope"
ACTIVITY = "Activity [MBq]"
DURATION = "Duration [h]"
SELF_SHIELDING = "Self shielding"    

# Legacy
APPLY_DECAY_CORRECTION = "Apply decay correction"
BIOLOGICAL_HALFLIFE = "Biological half-life [h]"
APPLY_BIOLOGICAL_DECAY = "Apply biological decay"
SELF_SHIELDING_PYSHIELD = 'Self shielding pyshield'

# display nuclear
PARENT = 'Parent'
ENERGY_KEV = 'Energy [keV]'
ABUNDANCE = 'Abundance [%]'


# New decay model
CLEARANCE = 'Clearance'
EMPTY_CLEARANCE = 'No Clearance'
APPLY_FRACTION1 = 'Apply Fraction 1'
APPLY_FRACTION2 = 'Apply Fraction 2'
DECAY_FRACTION1 = 'Fraction 1'
DECAY_FRACTION2 = 'Fraction 2'
HALFLIFE1 = 'Half Life 1 [h]'
HALFLIFE2 = 'Half Life 2 [h]'
SPLIT_FRACTION_TIME = 'Split Time [h]'
ENABLE_SPLIT_FRACTIONS = 'Split fractions by time'

# XRAY / CT
DAP = "DAP [Gycm2]"
DLP = "DLP [mGycm]"
CT_BODY_PART = "Body part"


NEW_PROJECT = 'New Project'

# Matplotlib
WALL_MARKER = 'WallMarker'
PIXEL_SIZE_MARKER = 'PixelSizeMarker'

MODEL_ITEMS = [SOURCES_CT, SOURCES_NM, SOURCES_XRAY, WALLS, SHIELDINGS,
               CRITICAL_POINTS]

SOURCES = [SOURCES_CT, SOURCES_NM, SOURCES_XRAY]

# GUI Toolbar
LABEL = 'Label'
ICON = 'Icon'
TOOILTIP = 'Tooltip'
CHECKABLE = 'Checkable'



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  7 11:25:57 2022

@author: Marcel
"""
from PyQt5.QtWidgets import (QPushButton, QLabel, 
                             QComboBox, QDoubleSpinBox, QApplication)
from PyQt5.QtGui import QIcon, QFont
import qtawesome                             

from pyrateshield.constants import CONSTANTS
from pyrateshield.model import DEFAULTS
from pyrateshield.model_sequences import Materials
from pyrateshield.labels import EMPTY_MATERIAL, MATERIALS

from pyrateshield.pyshield.calculator import attenuation_isotope, buildup_isotope, transmission_isotope, dose_rate
from pyrateshield.gui.item_views import EditValueWidget

from pyrateshield.radtracer.radtracer import TransmissionMCNP, MCNP_LOOKUP

class View(EditValueWidget):
    def create_widgets(self):
        
        bold_font = QFont()
        bold_font.setBold(True)

        self.isotope_list = QComboBox()
        self.material_list = QComboBox()
        self.thickness_value = QDoubleSpinBox() 

        self.pyshield_transmission = QLabel('')
        self.pyshield_attenuation = QLabel('')
        self.pyshield_buildup = QLabel('')
        
        self.pyshield_unshielded_doserate = QLabel('')
        self.radtracer_unshielded_doserate = QLabel('')
        
        self.pyshield_shielded_doserate = QLabel('')
        self.radtracer_shielded_doserate = QLabel('')
        
        self.radtracer_transmission = QLabel('')
        self.calculate_button = QPushButton('Calculate')
        self.close_button = QPushButton('Close')
    
    def create_layout(self):
        self.bold_font = QFont()
        self.bold_font.setBold(True)
        
        self.setWindowTitle('Calculator')
        icon = qtawesome.icon('mdi.calculator')
        self.setWindowIcon(QIcon(icon))
        
        row = 0
        
        self.layout.addWidget(QLabel('Isotope:'), row, 0)
        self.layout.addWidget(self.isotope_list, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Material:'), row, 0)
        self.layout.addWidget(self.material_list, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Thickness [cm]:'), row, 0)
        self.layout.addWidget(self.thickness_value, row, 1)
        
        row += 1
        label = QLabel('Unshielded h10 [uSv/h per MBq/m2]')
        label.setFont(self.bold_font)
        self.layout.addWidget(label, row, 0, 1, 2)    

        row += 1
        
        self.layout.addWidget(QLabel('Pyshield'))
        self.layout.addWidget(self.pyshield_unshielded_doserate, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Radtracer'))
        self.layout.addWidget(self.radtracer_unshielded_doserate, row, 1)
        
        
        row += 1
        
        label = QLabel('Shielding')
        label.setFont(self.bold_font)
        self.layout.addWidget(label, row, 0, 1, 2)

        row += 1 
        
        self.layout.addWidget(QLabel('Pyshield Attenuation:'), row, 0)
        self.layout.addWidget(self.pyshield_attenuation, row, 1)
        
        row += 1 
        
        self.layout.addWidget(QLabel('Pyshield Buildup:'), row, 0)
        self.layout.addWidget(self.pyshield_buildup, row, 1)
        
        row += 1 
        
        self.layout.addWidget(QLabel('Pyshield Transmission:'), row, 0)
        self.layout.addWidget(self.pyshield_transmission, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Radtracer Transmission:'), row, 0)
        self.layout.addWidget(self.radtracer_transmission, row, 1)

        row += 1
       
        label = QLabel('Shielded h10 [uSv/h per MBq/m2]')
        label.setFont(self.bold_font)
        self.layout.addWidget(label, row, 0, 1, 2)    

        row += 1
       
        self.layout.addWidget(QLabel('Pyshield'))
        self.layout.addWidget(self.pyshield_shielded_doserate, row, 1)
       
        row += 1
       
        self.layout.addWidget(QLabel('Radtracer'))
        self.layout.addWidget(self.radtracer_shielded_doserate, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.calculate_button, row, 0)
        self.layout.addWidget(self.close_button, row, 1)
        
        
        
        
        

class Controller:
    DISP_FORMAT = "{:.2e}"
    
    def __init__(self):
        self.view = View()
        self.materials = Materials.from_list(DEFAULTS[MATERIALS])
        self.view.material_list.setModel(self.materials)
        self.view.isotope_list.addItems(CONSTANTS.get_isotope_name_list())
        
        self.view.calculate_button.clicked.connect(self.calculate)
        self.view.close_button.clicked.connect(self.view.close)
    
    def calculate(self):
        isotope = self.view.isotope_list.currentText()
        material = self.materials.itemByName(self.view.material_list.currentText())
        thickness = self.view.thickness_value.value()
        
        if material.name != EMPTY_MATERIAL:
            pyshield_attenuation = attenuation_isotope(isotope, material, thickness)
            pyshield_buildup = buildup_isotope(isotope, material, thickness)
            pyshield_transmission = transmission_isotope(isotope, material, thickness)
            radtracer_transmission = TransmissionMCNP(MCNP_LOOKUP[isotope]['None']).get(material.name, thickness)
        else:
            pyshield_attenuation = 1
            pyshield_buildup = 1
            pyshield_transmission = 1
            radtracer_transmission = 1
        
        pyshield_unshielded_doserate = dose_rate(isotope)
        pyshield_shielded_doserate = pyshield_unshielded_doserate * pyshield_transmission
            
        radtracer_unshielded_doserate = MCNP_LOOKUP[isotope]['None']["h(10) [uSv/h per MBq/m^2]"]
        radtracer_shielded_doserate = radtracer_unshielded_doserate * radtracer_transmission        
        
        self.view.pyshield_attenuation.setText(self.to_str(pyshield_attenuation))
        self.view.pyshield_buildup.setText(self.to_str(pyshield_buildup))
        self.view.pyshield_transmission.setText(self.to_str(pyshield_transmission))
        self.view.pyshield_unshielded_doserate.setText(self.to_str(pyshield_unshielded_doserate))
        self.view.pyshield_shielded_doserate.setText(self.to_str(pyshield_shielded_doserate))
        
        self.view.radtracer_unshielded_doserate.setText(self.to_str(radtracer_unshielded_doserate))
        self.view.radtracer_shielded_doserate.setText(self.to_str(radtracer_shielded_doserate))
        self.view.radtracer_transmission.setText(self.to_str(radtracer_transmission))
        
        
    def to_str(self, value):
        return self.DISP_FORMAT.format(value)
        

if __name__ == "__main__":
    app = QApplication([])
    
    controller = Controller()
    window = controller.view
    window.show()    
    app.exec_()
    
    
    
    



import multiprocessing
from multiprocessing.sharedctypes import Value as mpValue
import ctypes
import queue
import time
import psutil
import os
from pyrateshield.constants import CONSTANTS
from pyrateshield import labels
from pyrateshield.logger import Logger, GLOBAL_LOG_LEVEL
from PyQt5.QtCore import QObject, pyqtSignal
from pyrateshield.radtracer import radtracer
from pyrateshield.pyshield.engine import Engine
import pandas as pd
from scipy.ndimage import zoom
import sys

NO_WORK = 0
DOSEMAP_WORK = 1
CRITICAL_POINT_WORK = 2


MAX_CORES = 8
LOG_LEVEL = Logger.LEVEL_DEBUG

UNSUPPORTED_PYSHIELD_ISOTOPES = ['Y-90']
CRITICAL_POINT_NAME         = 'Critical Point Name'
RADTRACER_DOSE              = 'RadTracer Dose [mSv]'
PYSHIELD_DOSE               = 'PyShield Dose [mSv]'
OCCUPANCY_FACTOR            = 'Occupancy factor'
RADTRACER_DOSE_CORRECTED    = 'RadTracer Dose corrected for occupancy [mSv]'
PYSHIELD_DOSE_CORRECTED     = 'PyShield Dose corrected for occupancy [mSv]'
SOURCE_NAME                 = 'Source Name'

COLUMNS = [CRITICAL_POINT_NAME, PYSHIELD_DOSE, RADTRACER_DOSE, 
           OCCUPANCY_FACTOR, PYSHIELD_DOSE_CORRECTED, RADTRACER_DOSE_CORRECTED]

SPLIT_COLUMNS = [CRITICAL_POINT_NAME, SOURCE_NAME, PYSHIELD_DOSE, 
                 RADTRACER_DOSE, OCCUPANCY_FACTOR, 
                 RADTRACER_DOSE_CORRECTED, PYSHIELD_DOSE_CORRECTED]

def resize_array(source_array, target_shape):
    # Need faster method for painting dosemap on top of the floorplan.
    # Obtain same size by interpolation
    zoom_factors = [float(t) / s for s, t in zip(source_array.shape, target_shape)]
    resized_array = zoom(source_array, zoom_factors, order=1)
    return resized_array

def is_supported_by_pyshield(src):
    supported = CONSTANTS.pyshield_supported_isotopes
    return src.label == labels.SOURCES_NM and src.isotope in supported


def is_supported_by_radtracer(src):
    supported = CONSTANTS.radtracer_supported_isotopes
    return src.label != labels.SOURCES_NM or src.isotope in supported


def get_radtracer_dosemap(source, project, pyshield_engine):
    if is_supported_by_radtracer(source):
        return radtracer.dosemap_single_source(source, project)
    elif is_supported_by_pyshield(source): # fallback to pyshield
        print(f'Source {source.name} not supported by radtracer, will use pyshield')   
        return pyshield_engine.source_dosemap(source)

def get_pyshield_dosemap(source, project, pyshield_engine):
    if is_supported_by_pyshield(source):    
        return pyshield_engine.source_dosemap(source)
    elif is_supported_by_radtracer(source): # fallback to radtracer
        # Xray will go through radtracer any how no need to warn
        if source.label == labels.SOURCES_NM:
            print(f'Source {source.name} not supported by pyshield, will use radtracer')
        return radtracer.dosemap_single_source(source, project)
    
def get_dosemap(source, project, pyshield_engine=None):
    if pyshield_engine is None:
        pyshield_engine = Engine.from_pyrateshield(project)
    # Pyshield set as engine
    if project.dosemap.engine == labels.PYSHIELD:
        return get_pyshield_dosemap(source, project, pyshield_engine=pyshield_engine)
    # Radtracer set as engine
    elif project.dosemap.engine == labels.RADTRACER:
        return get_radtracer_dosemap(source, project, pyshield_engine=pyshield_engine)

def empty_critical_point_format():
    return pd.DataFrame([{CRITICAL_POINT_NAME:      None,
                          SOURCE_NAME:              None,
                          RADTRACER_DOSE:           None,
                          PYSHIELD_DOSE:            None,
                          OCCUPANCY_FACTOR:         None,
                          RADTRACER_DOSE_CORRECTED: None,
                          PYSHIELD_DOSE_CORRECTED:  None
                          }])

def source_critical_point_format(source, 
                                 critical_point, 
                                 dose_radtracer, 
                                 dose_pyshield):
    
    factor = critical_point.occupancy_factor
    
    return {CRITICAL_POINT_NAME:        critical_point.name,
            SOURCE_NAME:                source.name,
            RADTRACER_DOSE:             dose_radtracer,
            PYSHIELD_DOSE:              dose_pyshield,
            OCCUPANCY_FACTOR:           factor,
            RADTRACER_DOSE_CORRECTED:   factor * dose_radtracer,
            PYSHIELD_DOSE_CORRECTED:    factor * dose_pyshield}
                          
def get_critical_point_for_source(project, critical_point, source,
                                  engine=labels.RADTRACER,
                                  pyshield_engine=None):
    
    if engine == labels.PYSHIELD and pyshield_engine is None:
        pyshield_engine = Engine.from_pyrateshield(project)
    
    def get_radtracer_dose(source):
        return radtracer.pointdose_single_source(
            critical_point.position, source, project)
    
    def get_pyshield_dose(source):
        return pyshield_engine.dose_at_point(
                critical_point.position, sources=[source])
    
    if not is_supported_by_pyshield(source)\
        and not is_supported_by_radtracer(source):
        raise RuntimeError('Unsupported source!')
        
    if engine == labels.PYSHIELD and is_supported_by_pyshield(source):
        dose = get_pyshield_dose(source)
    elif engine == labels.PYSHIELD:
        dose = get_radtracer_dose(source)
    elif engine == labels.RADTRACER and is_supported_by_radtracer(source):
        dose = get_radtracer_dose(source)
    elif engine == labels.RADTRACER:
        dose = get_pyshield_dose(source)
    return dose
    

def get_critical_point(project, critical_point, 
                       engine=labels.RADTRACER,
                       pyshield_engine=None,
                       sum_dose = True):
    
    sources = Dosemapper.project_sources(project)
    
    total_dose = {}
    for source in sources:
        dose = get_critical_point_for_source(project, critical_point, source, 
                                             engine=engine, 
                                             pyshield_engine=pyshield_engine)
        
        total_dose[(source.label, source.name)] = dose
    
    if sum_dose:
        total_dose = sum(total_dose.values())
        
    return total_dose


def get_formatted_critical_point_result(project, critical_point, source, 
                                        pyshield_engine=None):
    
    dose_pyshield = get_critical_point_for_source(
        project, critical_point, source,
        engine=labels.PYSHIELD,
        pyshield_engine=pyshield_engine)
        
    dose_radtracer = get_critical_point_for_source(
        project, critical_point, source,
        engine=labels.RADTRACER,
        pyshield_engine=pyshield_engine)
   
    formatted = source_critical_point_format(
        source, critical_point, 
        dose_radtracer=dose_radtracer,
        dose_pyshield=dose_pyshield)
    
    return formatted
    
class Worker(multiprocessing.Process):
    def __init__(self, project_queue, work_queue, 
                 source_queue, workt_type, update_flag, stop_flag):
        multiprocessing.Process.__init__(self)
        
   
        
        self.project_queue = project_queue
        
        self.source_queue = source_queue
        
        self.work_type = workt_type
        
        self.work_queue = work_queue
       
        self.update_flag = update_flag
        
        self.stop_flag = stop_flag      
        
    
    def update_project(self):
        
        if self.update_flag.value:        
        
            self.project = self.project_queue.get(timeout=2)
      
            self.pyshield_engine = Engine.from_pyrateshield(self.project)
        
        self.update_flag.value = False
        
    def run(self):
        while not self.stop_flag.value:  
            work_type = self.work_type.value
            if self.update_flag.value:
                self.update_project()
                
            if work_type == NO_WORK:
                time.sleep(0.1)
               
            elif work_type > NO_WORK and not self.source_queue.empty(): 
                try:
                    sources = self.source_queue.get(timeout=0.1)
                    # work type changed during time out 
                    if work_type != self.work_type.value:
                        self.source_queue.put(sources)
                        continue
                except queue.Empty:
                    continue

                
                for source in sources:
                    if work_type == DOSEMAP_WORK:

                        self.work_queue.put(get_dosemap(source, self.project))
                        
                    elif work_type == CRITICAL_POINT_WORK:
                        self.work_queue.put(self.get_critical_points(source))
                    else:
                        raise RuntimeError
                
        
            
    def get_critical_points(self, source):
        results = []
        for critical_point in self.project.critical_points:
            result = get_formatted_critical_point_result(self.project, 
                                                         critical_point, 
                                                         source,
                                                pyshield_engine=self.pyshield_engine)
            results.append(result)
        return results
    

   

class Dosemapper(QObject):
    _workers = None
    _work_queue = None
    _max_cpus = MAX_CORES
    _source_queue = None
    _cpus = None
    progress = pyqtSignal(float)
    calculation_time = pyqtSignal(float)
    interpolate = pyqtSignal(float)
    def __init__(self, multi_cpu=True):
        super().__init__()
        if not multi_cpu:
            self.set_cpus(1)

    
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._stop_workers(self._workers)
       
    def source_queue(self):
        if self._source_queue is None:
            self._source_queue = multiprocessing.Queue()
        return self._source_queue
    
    def work_queue(self):
        if self._work_queue is None:
            self._work_queue  = multiprocessing.Queue()
        return self._work_queue
    
    def set_cpus(self, cpus=None):
        if self._cpus != cpus:
            self._cpus = cpus
            
    def cpus(self):
        if self._cpus is None:
            max_cpus = psutil.cpu_count(logical=False)
            self._cpus = min(max_cpus, self._max_cpus)
            self._cpus = max(self._cpus, 1)
        return self._cpus
    
    def update_workers(self, project=None, work_type=None):
        
        while len(self.workers()) < self.cpus():
            self.workers().append(self._new_worker())
            
        while len(self.workers()) > self.cpus():
           
            self._stop_worker(self.workers()[0])
            
            self._workers.remove(self._workers[0])
        
        while any(w.update_flag.value for w in self.workers()):
            time.sleep(0.1)
        
        if project is not None:
            for w in self.workers():
                w.project_queue.put(project)
                w.work_type.value = work_type if work_type is not None else 0
                w.update_flag.value = True
        
        if work_type is not None:
            for w in self.workers():
                w.work_type.value = work_type
            
        return self.workers()
    
    def workers(self):
        if self._workers is None:
            self._workers = []
        return self._workers
    
    def _new_worker(self):
        project_queue   = multiprocessing.Queue()
        stop_flag       = mpValue(ctypes.c_bool, False)
        update_flag     = mpValue(ctypes.c_bool, False)
        work_type       = mpValue(ctypes.c_int8, NO_WORK)
        args = (
            project_queue,
            self.work_queue(),
            self.source_queue(),
            work_type,
            update_flag,
            stop_flag,
        )
        p = Worker(*args)            
        p.start()
        return p

    def _stop_worker(self, worker):
        worker.stop_flag.value = True
        worker.join()
        time.sleep(0.1)
        if worker.is_alive():
            worker.close()
    
    def _stop_workers(self, workers):
        while workers is not None and len(workers) > 0:
            self._stop_worker(workers[0])
            workers.remove(workers[0])

   
    @staticmethod
    def project_sources(project):
        sources = list(project.sources_nm) + list(project.sources_ct)\
            + list(project.sources_xray)
        return [source for source in sources if source.enabled]
        
        
   
    def get_single_cpu_dosemap(self, project, sources=None):
        # DEBUGGING Purpuses
    
        start = time.time()
        dosemap = None        
        update_dosemap = lambda dm: dm if dosemap is None else (dosemap + dm)
            
        
        sources = Dosemapper.project_sources(project)
        
        pyshield_engine = Engine.from_pyrateshield(project)
         
        for i, source in enumerate(sources):
            source_dosemap = get_dosemap(source, project, pyshield_engine=pyshield_engine)
            self.progress.emit(i/len(sources))
            
            dosemap = update_dosemap(source_dosemap)
        
        stop = time.time()
        self.progress.emit(1)
        self.calculation_time.emit(stop-start)
        
        if project.dosemap_style.interpolate:
            self.interpolate.emit(-1)
            start = time.time()
            dosemap = resize_array(dosemap, project.floorplan.image.shape)
            stop = time.time()
            
            self.interpolate.emit(stop-start)
            
            
            
        return dosemap
 
    def get_critical_point(self, project, sources=None):
        if sources is None:
            sources = self.project_sources(project)
        
        src_pos_dct = {}

        for source in sources:
            # Make a dictionary of source_positions and group the corresponding sources
            # I.e.: { (x,y,z): [src1, src2, src3 ], (x,y,z): [src4, src5, ... ] }
            src_pos_dct.setdefault(tuple(source.position), []).append(source)
            
        sources_grouped = [srcs for srcs in src_pos_dct.values()]
        
        self.update_project(project)
        
        for grp in sources_grouped:
            self.source_queue().put(grp)
            
        
    def get_dosemap(self, project, sources=None):
        #print(f'Using engine: {project.dosemap.engine}')
        
        
        if self.cpus() == 1 or not project.dosemap_style.multi_cpu:
            return self.get_single_cpu_dosemap(project, sources)
        
        start = time.time()
      
        self.update_workers(project=project)
        
        
        
        if sources is None:
            sources = self.project_sources(project)
     
        
        # Check if there is any work to be done
        if not len(sources):
            return None

        
        # Sources with the same position are processed by the same worker.
        # This allows PyShield to benefit from caching.
        # MS: Group sources by default won't affect radtracer
            
        src_pos_dct = {}

        for source in sources:
            # Make a dictionary of source_positions and group the corresponding sources
            # I.e.: { (x,y,z): [src1, src2, src3 ], (x,y,z): [src4, src5, ... ] }
            src_pos_dct.setdefault(tuple(source.position), []).append(source)
            
        sources_grouped = [srcs for srcs in src_pos_dct.values()]

        
        self.update_workers(work_type=DOSEMAP_WORK)
        
        for grp in sources_grouped:
            self.source_queue().put(grp)
            
        
        # Collect the results. Make sure to get as many results from the queue
        # as the number of sources
        nr_of_sources = sum(len(grp) for grp in sources_grouped)
                
        dosemap = None        
        update_dosemap = lambda dm: dm if dosemap is None else (dosemap + dm)
       

        self.progress.emit(0)
        for i in range(nr_of_sources):
            dosemap = update_dosemap(self.work_queue().get())
            self.progress.emit(i/nr_of_sources)
        
        stop = time.time()
        
        self.progress.emit(1)
        self.calculation_time.emit(stop-start)
        self.update_workers(work_type=NO_WORK)
        
        if not self.work_queue().empty(): raise RuntimeError()
        
        if project.dosemap_style.interpolate:
            self.interpolate.emit(-1)
            start = time.time()
            dosemap = resize_array(dosemap, project.floorplan.image.shape)
            stop = time.time()
            
            self.interpolate.emit(stop-start)
        return dosemap        
    
    def get_critical_points(self, project, sum_sources=True):
    
        
        
        self.update_workers(project=project)
       
        
        sources = self.project_sources(project)
        
        
        if not len(sources):
            return empty_critical_point_format()
       
        
        if self.cpus() == 1 or not project.dosemap_style.multi_cpu:
            start = time.time()
            results = []
            pyshield_engine = Engine.from_pyrateshield(project)
            self.progress.emit(0)
            for i, source in enumerate(sources): 
                for critical_point in project.critical_points:
                    result = get_formatted_critical_point_result(project, critical_point, source,
                                                                 pyshield_engine=pyshield_engine)
                    results += [result]
                    self.progress.emit(i/len(sources))
            stop = time.time()
            self.calculation_time.emit(stop-start)
            self.progress.emit(1)
            
        else:
            start = time.time()
            
            src_pos_dct = {}
            
            for source in sources:
                # Make a dictionary of source_positions and group the corresponding sources
                # I.e.: { (x,y,z): [src1, src2, src3 ], (x,y,z): [src4, src5, ... ] }
                src_pos_dct.setdefault(tuple(source.position), []).append(source)
           
            sources_grouped = [srcs for srcs in src_pos_dct.values()]
            
            
            
            for grp in sources_grouped:
                self.source_queue().put(grp)
            
            
            
            nr_of_sources = sum(len(grp) for grp in sources_grouped)
            
            
           
            
            
            results = None
            
            update_results = lambda result: result if results is None else (results+result)
          
            self.progress.emit(0)
            
            self.update_workers(work_type=CRITICAL_POINT_WORK)
           
            for i in range(nr_of_sources):
                result = self.work_queue().get()
                
                results = update_results( result)    
                
                self.progress.emit(i/nr_of_sources)
            stop = time.time()
            self.calculation_time.emit(stop-start)
            self.progress.emit(1)
            self.update_workers(work_type=NO_WORK)
            
        report = pd.DataFrame(results)[SPLIT_COLUMNS]
        
        if sum_sources:
            aggfunc = {RADTRACER_DOSE: sum, RADTRACER_DOSE_CORRECTED: sum,
                       PYSHIELD_DOSE: sum, PYSHIELD_DOSE_CORRECTED: sum,
                       OCCUPANCY_FACTOR: max}
                
            report = pd.pivot_table(report, index=CRITICAL_POINT_NAME, 
                                            aggfunc=aggfunc)
            
            # preserve order
            pnames = [crit_point.name for crit_point in project.critical_points\
                      if crit_point.enabled]
                
            report = report.reindex(pnames)
            
            report = report.reset_index()
           
            report = report[COLUMNS]
            
       
        
        return report
            
    
    # @staticmethod
    # def get_sources_critical_points(project, sum_sources=True):
    #     sources = list(project.sources_nm) + list(project.sources_ct) + list(project.sources_xray)
        
    #     # Exclude sources which are not set to Enabled
    #     sources = [src for src in sources if src.enabled]
        
    #     # Exclude points which are not set to Enabled
    #     crit_points = [crp for crp in project.critical_points if crp.enabled]
                
    #     if len(crit_points) == 0 or len(sources) == 0:            
    #         return Dosemapper.empty_critical_point_result()
        
    #     reports = []        
        
    #     pyshield_engine = Engine.from_pyrateshield(project)
        
    #     for crit_point in crit_points:            
    #         for source in sources:
    #             if is_supported_by_radtracer(source):
    #                 dose_radtracer = radtracer.pointdose_single_source(
    #                     crit_point.position, source, project)
    #             else:
    #                 # At least one isotope not supported by radtracer
    #                 dose_radtracer = None
                
    #             if is_supported_by_pyshield(source):
    #                 dose_pyshield = pyshield_engine.dose_at_point(crit_point.position,
    #                                                      sources=[source])
    #             else:
    #                 dose_pyshield = None
                    
    #             if dose_radtracer is None and dose_pyshield is None:
    #                 raise RuntimeError
    #             elif dose_radtracer is None:
    #                 dose_radtracer = dose_pyshield
    #             elif dose_pyshield is None:
    #                 dose_pyshield = dose_radtracer
                    
                
    #             reports += [Dosemapper.get_source_critical_point_result(
    #                 source, crit_point, dose_radtracer, dose_pyshield)]
    #     report = pd.DataFrame(reports)[SPLIT_COLUMNS]
        
    #     if sum_sources:
    #         aggfunc = {RADTRACER_DOSE: sum, RADTRACER_DOSE_CORRECTED: sum,
    #                    PYSHIELD_DOSE: sum, PYSHIELD_DOSE_CORRECTED: sum,
    #                    OCCUPANCY_FACTOR: max}
                
    #             report = pd.pivot_table(report, index=CRITICAL_POINT_NAME, 
    #                                             aggfunc=aggfunc)
                
    #             # preserve order
    #             pnames = [crit_point.name for crit_point in project.critical_points\
    #                       if crit_point.enabled]
                    
    #             report = summed_report.reindex(pnames)
                
    #             report = summed_report.reset_index()
               
    #             report = summed_report[COLUMNS]
                
        
    #     return report
        
    # @staticmethod
    # def get_critical_points(project):
        
        
    #     if len(project.sources_nm) + len(project.sources_ct) + len(project.sources_xray) == 0\
    #         or len(project.critical_points) == 0:
    #             return Dosemapper.empty_critical_point_result()
            
    #     report = Dosemapper.get_sources_critical_points(project)
        
    #     aggfunc = {RADTRACER_DOSE: sum, RADTRACER_DOSE_CORRECTED: sum,
    #                PYSHIELD_DOSE: sum, PYSHIELD_DOSE_CORRECTED: sum,
    #                OCCUPANCY_FACTOR: max}
        
    #     summed_report = pd.pivot_table(report, index=CRITICAL_POINT_NAME, 
    #                                    aggfunc=aggfunc)
        
    #     # preserve order
    #     pnames = [crit_point.name for crit_point in project.critical_points\
    #               if crit_point.enabled]
            
    #     summed_report = summed_report.reindex(pnames)
        
    #     summed_report = summed_report.reset_index()
       
    #     summed_report = summed_report[COLUMNS]
    #     return summed_report

    
        
            
if __name__ == "__main__":
    from pyrateshield.model import Model
    import timeit
    import time
    #model = Model.load_from_project_file('/Users/marcel/git/pyrateshield/example_projects/LargeProject/large_project.psp')
    #model = Model.load_from_project_file('../example_projects/LargeProject/project.zip')
    model = Model.load_from_project_file('../example_projects/SmallProject/project.zip')
    #model = Model.load_from_project_file('../example_projects/Stabin/Tb-161.zip')
    #model = Model.load_from_project_file('../example_projects/Lu-177.psp')
    model.dosemap.grid_matrix_size = 120
    
    
    
        
    with Dosemapper(multi_cpu=True) as dm:   
        dm.progress.connect(print)
        model.dosemap.engine = labels.PYSHIELD
        #dm = dm.get_dosemap(model)
        #print("PyShield", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
        print("PyShield", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
    
        #time.sleep(1) 
        #print()
        
        model.dosemap.engine = labels.RADTRACER
        #print("Radtracer", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
        #print("Radtracer", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
        #results = dm.get_critical_points(model)
        print(timeit.timeit(lambda: dm.get_critical_points(model), number=1))

    # from pyshield import Sources
    # sources = Sources.from_pyrateshield(model)
    # print("Pyshield single CPU:", end=" ", flush=True)
    # print(timeit.timeit(lambda: sources.get_dosemap(), number=1) )
    
    
    
    




# -*- coding: utf-8 -*-
"""
Created on Mon Feb 21 14:41:18 2022

@author: 757021
"""
from PyQt5.QtCore import pyqtSignal, QAbstractListModel, Qt, QModelIndex
from pyrateshield import model_items, labels
import qtawesome as qta

import yaml


class ModelItemSequence(QAbstractListModel):
    nameChanged = pyqtSignal(str, str)
    valueChanged = pyqtSignal(object)
    
    default_item_name = None
    empty_item_name = None
    
    def __init__(self, *args, items=None,  **kwargs):
        QAbstractListModel.__init__(self, *args, **kwargs)
        
        items = items or []
        
        if self.empty_item_name is not None:
            names = [item.name for item in items]
            if self.empty_item_name not in names:
                items = [self.item_class(name=self.empty_item_name), *items]
       
        self.items = []
        for item in items:
            self.addItem(item)
    
    # Tricky stuff to make the object pickable
    def __getstate__(self):
        return {'items': [item.to_dict() for item in self.items]}
    
    def __setstate__(self, dct):        
        self.__dict__ = dct
        self.items = [self.item_class.from_dict(item) for item in self.items]
    # End tricky stuff
        
    def __getitem__(self, index):
        return self.items[index]
    
    def __len__(self):
        return len(self.items)
    
    def __str__(self):
        return yaml.dump(self.to_list())
    
    def __repr__(self):
        return self.__str__()
    
    
    def data(self, index, role):

        row = index.row()
        nrows = len(self)
        
        
        if row >= nrows:
            row = nrows - 1
            
        if role == Qt.DisplayRole:

            if row >= 0:
                item = self[row]
            else:
                item = None
                
            
            if hasattr(item, 'name'):
                return item.name
            elif item is not None:
                return row
            
    def deleteItem(self, item):
        if hasattr(item, 'name') and item.name == self.empty_item_name:
            return False
       
        index = self.items.index(item)
        index = self.createIndex(self.items.index(item), 0)
        
        if not index.isValid() or not (0 <= index.row() < len(self.items)):
            return False

        item = self.itemAtIndex(index)
      
        self.beginRemoveRows( self.createIndex(index.row(), 0), index.row(), 
                             index.row())

        del self.items[index.row()]
     
        self.endRemoveRows()
        
        

        return True
    
    def addItem(self, item):
    
        self.beginInsertRows(QModelIndex(), len(self.items), len(self.items))
        item = self.checkName(item)
        self.items.append(item)     
        item.setParent(self)
        item.update_event.connect(self.emitDataChanged)
        self.endInsertRows()
        
    def addItems(self, items):
        self.beginInsertRows(QModelIndex(), len(self.items), len(self.items))
        for item in self.items:
            item.setParent(self)
            item.update_event.connect(self.emitDataChanged)
            self.items.append(item)
        self.endInsertRows()
          
    def emitDataChanged(self, event_data):
        item, label, old_value, value = event_data
        if label == labels.NAME:
            self.nameChanged.emit(old_value, value)
        self.valueChanged.emit(event_data)
        index = self.indexForItem(item)
        if index is not None:
            self.dataChanged.emit(index, index)
        
    def itemByName(self, name):
        if name not in self.itemNames():
            return None
        else:
            return self.items[self.itemNames().index(name)]
        
    def itemNames(self):
        return [item.name for item in self.items]
        
        
    def indexForItem(self, item):
        if item not in self.items:
            return None
        else:
            index = self.items.index(item)
            return self.createIndex(index, index)

    def rowCount(self, index):
        return len(self.items)
    
    def itemAtIndex(self, index):
        row = index if isinstance(index, int) else index.row()
        if len(self) == 0 or row >= len(self):
            return None
        if isinstance(index, int):
            return self.items[index]        
        elif isinstance(index, QModelIndex):
            return self.items[index.row()]
    
    def checkName(self, item):
        if hasattr(item, 'name'):
            names = [item.name for item in self.items]
            if item.name in names:
                item.name = self.get_new_name()
        return item
    
    def get_new_name(self):
        # Just add a counter to the defualt name for the new name. If new name 
        # already in self.names then the counter is increased until the
        # new name is not already in self.names
        new_name =  self.item_class.default_name + ' 1'
        i = 1
        while new_name in [item.name for item in self.items]:
            i += 1
            new_name = self.item_class.default_name + ' ' + str(i)
        return new_name  
    
    @classmethod
    def from_list(cls, items):
        return cls(items=[cls.item_class.from_dict(item) for item in items])
    
    
    def to_list(self):
        return [item.to_dict() for item in self.items]

class CriticalPoints(ModelItemSequence):
    item_class = model_items.CriticalPoint
    default_item_name = 'critical point'

class Walls(ModelItemSequence):
    item_class = model_items.Wall
    shieldingChanged = pyqtSignal(object)
        
    def closestVertexToPixelPoint(self, position, exclude_wall=None):
        distance = float('Inf')
        closest_wall = None
        closest_index = None
        for wall in self:
            if wall is exclude_wall: continue
            index, vertex, di = wall.closestVertexToPixelPoint(position)
            if di < distance:
                closest_wall = wall
                closest_index = index
                distance = di
        return closest_wall, closest_index, distance
                
    def emitDataChanged(self, event_data):     
        super().emitDataChanged(event_data)
        item, label, old_value, value = event_data
        if label == labels.SHIELDING:                     
            self.shieldingChanged.emit(item)
            
    def simplifiedWalls(self):
        walls = []
        for item in self:            
            for i in range(0, len(item.vertices)-1):
                wall = item.from_dict(item.to_dict()) # copy 
                wall.vertices = item.vertices[i: i + 2]
                walls += [wall]
            if item.isClosed():
                wall = item.from_dict(item.to_dict())
                wall.vertices = [wall.vertices[-1], wall.vertices[0]]
                walls += [wall]
        return walls
                                                
            
class SourcesNM(ModelItemSequence):
    item_class = model_items.SourceNM

class SourcesXray(ModelItemSequence):
    item_class = model_items.SourceXray

class SourcesCT(ModelItemSequence):
    item_class = model_items.SourceCT
 
class Clearances(ModelItemSequence):
    item_class = model_items.Clearance
    
    empty_item_name = labels.EMPTY_CLEARANCE
    
    def items_in_use(self):
        item_names = [source.clearance for source in self.parent().sources_nm]
        items =  [self.itemByName(name) for name in item_names]
        return items
        
    
class Materials(ModelItemSequence):
    item_class = model_items.Material
    empty_item_name = labels.EMPTY_MATERIAL
    
    def items_in_use(self):
    
        item_names = []
        for shielding in self.parent().shieldings:
            item_names += [shielding.materials[0][0]]
            if len(shielding.materials) > 1:
                item_names += [shielding.materials[1][0]]
        item_names = list(set(item_names))
        items = [self.itemByName(name) for name in item_names]
        return items
    
class Shieldings(ModelItemSequence):
    item_class = model_items.Shielding
    styleChanged = pyqtSignal(object)
    empty_item_name = labels.EMPTY_SHIELDING

    def emitDataChanged(self, event_data):
        super().emitDataChanged(event_data)
        item, label, old_value, value = event_data
        if label in (labels.COLOR, labels.LINEWIDTH):                     
            self.styleChanged.emit(item)
            
    def items_in_use(self):    
        item_names = [wall.shielding for wall in self.parent().walls]
        items =  [self.itemByName(name) for name in item_names]
        return items
    
    def data(self, index, role):
        if role == Qt.DecorationRole:
            item = self.itemAtIndex(index)
            if item is not None:
                return qta.icon('mdi.wall', color=item.color)
        else:
            return super().data(index, role)
        
    
    
class Rulers(ModelItemSequence):
    item_class = model_items.Ruler


import pandas as pd

from PyQt5.QtWidgets  import QFileDialog
from PyQt5 import QtCore



class PandasModel(QtCore.QAbstractTableModel):
    NUMBER_OF_DIGITS = 2
    def __init__(self, data, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent)
        self.original_data = data
        data = data.copy()
        
        self._data = data
        
        for col in data.columns:
            try:
                self._data[col] = self._data[col].apply(self.precision_round)
            except:
                pass
            
        self._cols = self._data.columns

    def rowCount(self, parent=None):
        return len(self._data.values)

    def columnCount(self, parent=None):
        return self._data.columns.size

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                return QtCore.QVariant(str(
                    self._data.iloc[index.row()][index.column()]))
        return QtCore.QVariant()
    
    def headerData(self, p_int, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self._cols[p_int]
            elif orientation == QtCore.Qt.Vertical:
                return p_int
        return None
    
    def precision_round(self, number, digits=None):
        if digits is None:
            digits = self.NUMBER_OF_DIGITS
            
        power = "{:e}".format(number).split('e')[1]
        return round(number, -(int(power) - digits))


def ask_new_file(title):
    extensions = "Excel files (*.xlsx)"
    filedialog = QFileDialog()
    filedialog.setFileMode(QFileDialog.ExistingFile)
    file = str(QFileDialog.getSaveFileName(filedialog, title, "",
                                           extensions)[0])
    
    return file


class CriticalPointReportController:
    _table_model = None

    def __init__(self, view=None, model=None, controller=None, dosemapper=None):
        self.dosemapper = dosemapper
        self.view = view
        self.model = model
        self.controller = controller
        self.view.save_critcial_point_button.setEnabled(False)
        
        callback = self.calculate_critical_points
        button = self.view.critical_point_button
        button.clicked.connect(callback)
        
        callback = self.save_critical_points
        button = self.view.save_critcial_point_button
        button.clicked.connect(callback)
        
    def set_model(self, model):
        self.model = model
        self.clear()
    
    def clear(self):
        if self.table_model is not None:
            self.table_model = PandasModel(pd.DataFrame())
            self.view.table_view.setModel(self.table_model)
        
    
    @property
    def table_model(self):
        return self._table_model
    
    @table_model.setter
    def table_model(self, table_model):
        if table_model is None:
            self.view.save_critcial_point_button.setEnabled(False)
        else:
            self.view.save_critcial_point_button.setEnabled(True)
        self._table_model = table_model    
        
    def calculate_critical_points(self):

        sum_sources = not self.view.source_checkbox.isChecked()
        pd_report = self.dosemapper.get_critical_points(self.model, 
                                                        sum_sources=sum_sources)
       
            
        self.table_model = PandasModel(pd_report)        
        self.view.table_view.setModel(self.table_model)
        
    def save_critical_points(self):
        file = ask_new_file('Save to Excel')
        if file != '':
            self.table_model.original_data.to_excel(file, engine='xlsxwriter')
            
            
        
        

        
      
